import pandas as pd

from dbt.adapters.base import BaseRelation
from dbt.adapters.base.connections import AdapterResponse

from dbt.adapters.fal_experimental.adapter_support import new_connection

# [snowflake] extras dependencies
import snowflake.connector as snowflake

from dbt.adapters.snowflake import SnowflakeAdapter, SnowflakeConnectionManager


def read_relation_as_df(
    adapter: SnowflakeAdapter, relation: BaseRelation
) -> pd.DataFrame:
    sql = f"SELECT * FROM {relation}"

    assert adapter.type() == "snowflake"

    with new_connection(adapter, "fal-snowflake:read_relation_as_df") as conn:
        handle: snowflake.SnowflakeConnection = conn.handle
        cur = handle.cursor()

        cur.execute(sql)
        df: pd.DataFrame = cur.fetch_pandas_all()

        # HACK: manually parse ARRAY and VARIANT since they are returned as strings right now
        # Related issue: https://github.com/snowflakedb/snowflake-connector-python/issues/544
        for desc in cur.description:
            # 5=VARIANT, 10=ARRAY -- https://docs.snowflake.com/en/user-guide/python-connector-api.html#type-codes
            if desc.type_code in [5, 10]:
                import json

                df[desc.name] = df[desc.name].map(lambda v: json.loads(v))

        return df


def write_df_to_relation(
    adapter: SnowflakeAdapter,
    data: pd.DataFrame,
    relation: BaseRelation,
) -> AdapterResponse:
    import snowflake.connector.pandas_tools as snowflake_pandas

    assert adapter.type() == "snowflake"

    database: str = relation.database  # type: ignore
    schema: str = relation.schema  # type: ignore
    table: str = relation.identifier  # type: ignore

    with new_connection(adapter, "fal-snowflake:write_df_to_relation") as conn:
        connection_manager: SnowflakeConnectionManager = adapter.connections  # type: ignore
        handle: snowflake.SnowflakeConnection = conn.handle

        with connection_manager.exception_handler("LOAD TABLE"):
            success, _, num_rows, output = snowflake_pandas.write_pandas(
                handle,
                data,
                table_name=table,
                database=database,
                schema=schema,
                overwrite=True,  # TODO: This helps when table schema changes, but it is not atomic
                auto_create_table=True,
                quote_identifiers=False,
            )
            if not success:
                # In case the failure does not raise by itself
                # I have not been able to reproduce such a case
                from dbt.exceptions import DbtDatabaseError

                raise DbtDatabaseError(output)

            # TODO: better AdapterResponse
            return AdapterResponse(str(output[0][1]), rows_affected=num_rows)



from collections import defaultdict
import os.path
from dataclasses import dataclass, field
from typing import (
    Dict,
    Iterable,
    List,
    Any,
    Optional,
    Tuple,
    Sequence,
    TYPE_CHECKING,
)
from pathlib import Path
from deprecation import deprecated

import fal.dbt.integration.version as version

from dbt.cli.resolvers import default_profiles_dir
from dbt.cli.main import dbtRunner, dbtRunnerResult

from dbt.contracts.graph.nodes import (
    SourceDefinition,
    TestMetadata,
    GenericTestNode,
    SingularTestNode,
)
from dbt.contracts.graph.nodes import ManifestNode

from dbt.contracts.graph.manifest import (
    Manifest,
    MaybeNonSource,
    MaybeParsedSource,
    Disabled,
)

from dbt.node_types import NodeType
from dbt.contracts.connection import AdapterResponse
from dbt.contracts.results import (
    RunResultsArtifact,
    RunResultOutput,
    NodeStatus,
    FreshnessExecutionResultArtifact,
    FreshnessNodeOutput,
)
from dbt.task.compile import CompileTask

from . import parse
from . import lib
from . import version

from fal.dbt.feature_store.feature import Feature

import pandas as pd

from fal.dbt.telemetry import telemetry
from fal.dbt.utils import has_side_effects

if TYPE_CHECKING:
    from fal.dbt.fal_script import Hook, TimingType
    from fal.dbt.packages.environments import BaseEnvironment


class FalGeneralException(Exception):
    pass


FAL = "fal"


@dataclass
class _DbtNode:
    node: Any = field(repr=False)
    _status: str = field(default=NodeStatus.Skipped.value)

    @property
    def name(self) -> str:
        return self.node.name

    @property
    def unique_id(self) -> str:
        return self.node.unique_id

    def _get_status(self):
        return self._status

    def _set_status(self, status: str):
        self._status = status

    status = property(_get_status, _set_status)


@dataclass
class DbtTest(_DbtNode):
    model_ids: List[str] = field(init=False, default_factory=list)
    source_ids: List[str] = field(init=False, default_factory=list)

    @classmethod
    def init(cls, node):
        if node.resource_type == NodeType.Test:
            if isinstance(node, GenericTestNode):
                test = DbtGenericTest(node=node)
            elif isinstance(node, SingularTestNode):
                test = DbtSingularTest(node=node)
            else:
                raise ValueError(f"Unexpected test class {node.__class__.__name__}")

            for dep in test.node.depends_on.nodes:
                if dep.startswith("model."):
                    test.model_ids.append(dep)
                if dep.startswith("source."):
                    test.source_ids.append(dep)

            return test
        else:
            raise TypeError(
                f"Initialized DbtTest with node of type {node.resource_type}"
            )


@dataclass
class DbtGenericTest(DbtTest):
    column: Optional[str] = field(init=False)

    def __post_init__(self):
        assert isinstance(self.node, GenericTestNode)
        self.column = self.node.column_name

        # Column name might be stored in test_metadata
        if not self.column and self.node.test_metadata.kwargs.get("column_name"):
            self.column = self.node.test_metadata.kwargs.get("column_name")

    @property
    def source_id(self):
        if self.source_ids:
            return self.source_ids[0]

    @property
    def model_id(self):
        if self.model_ids:
            return self.model_ids[0]

    # TODO: Deprecate?
    @property
    def source(self):
        if self.source_id:
            parts = self.source_id.split(".")
            return parts[-2], parts[-1]

    # TODO: Deprecate?
    @property
    def model(self):
        if self.model_id:
            # TODO: handle package models
            parts = self.model_id.split(".")
            return parts[-1]

    @property
    def name(self) -> str:
        metadata: TestMetadata = self.node.test_metadata
        return metadata.name


@dataclass
class DbtSingularTest(DbtTest):
    def __post_init__(self):
        assert isinstance(self.node, SingularTestNode)


@dataclass
class _DbtTestableNode(_DbtNode):
    # TODO: should this include singular tests that ref to this node?
    tests: List[DbtGenericTest] = field(default_factory=list)

    def _get_status(self):
        if self._status == NodeStatus.Skipped and any(
            test.status != NodeStatus.Skipped for test in self.tests
        ):
            return "tested"
        else:
            return self._status

    status = property(_get_status, _DbtNode._set_status)


@dataclass
class DbtSource(_DbtTestableNode):
    freshness: Optional[FreshnessNodeOutput] = field(default=None)

    def __repr__(self):
        attrs = ["name", "table_name", "tests", "status"]
        props = ", ".join([f"{item}={repr(getattr(self, item))}" for item in attrs])
        return f"DbtSource({props})"

    @property
    def meta(self):
        return self.node.meta

    @property
    def table_name(self) -> str:
        return self.node.name

    @property
    def name(self) -> str:
        return self.node.source_name


@dataclass
class DbtModel(_DbtTestableNode):
    python_model: Optional[Path] = field(default=None)

    _adapter_response: Optional[AdapterResponse] = field(default=None)

    def __repr__(self):
        attrs = ["name", "alias", "unique_id", "columns", "tests", "status"]
        props = ", ".join([f"{item}={repr(getattr(self, item))}" for item in attrs])
        return f"DbtModel({props})"

    @property
    def columns(self):
        return self.node.columns

    @property
    def alias(self):
        return self.node.alias

    @property
    def meta(self):
        return self.node.meta

    @property
    def is_incremental(self):
        return self.node.config.materialized == "incremental"

    def _get_adapter_response(self):
        return self._adapter_response

    def _set_adapter_response(self, adapter_response: Optional[dict]):
        self._adapter_response = (
            AdapterResponse.from_dict(adapter_response) if adapter_response else None
        )

    adapter_response = property(_get_adapter_response, _set_adapter_response)

    def __hash__(self) -> int:
        return self.unique_id.__hash__()

    def get_depends_on_nodes(self) -> List[str]:
        return self.node.depends_on_nodes

    def get_hooks(
        self,
        hook_type: "TimingType",
    ) -> List["Hook"]:
        from fal.dbt.fal_script import create_hook, TimingType

        meta = self.meta or {}

        keyword_dict = meta.get(FAL) or {}
        if not isinstance(keyword_dict, dict):
            return []

        if hook_type == TimingType.PRE:
            hook_key = "pre-hook"
        elif hook_type == TimingType.POST:
            hook_key = "post-hook"
        else:
            raise ValueError(f"Unexpected hook type {hook_type}")

        raw_hooks = keyword_dict.get(hook_key) or []
        if not isinstance(raw_hooks, list):
            return []

        return [
            create_hook(raw_hook, default_environment_name=self.environment_name)
            for raw_hook in raw_hooks
        ]

    def get_scripts(self, *, before: bool) -> List[str]:
        # sometimes properties can *be* there and still be None
        meta = self.meta or {}

        keyword_dict = meta.get(FAL) or {}
        if not isinstance(keyword_dict, dict):
            return []

        scripts_node = keyword_dict.get("scripts") or []
        if not scripts_node:
            return []

        if isinstance(scripts_node, list):
            if before:
                return []
            else:
                return scripts_node

        if not isinstance(scripts_node, dict):
            return []

        if before:
            return scripts_node.get("before") or []
        else:
            return scripts_node.get("after") or []

    @property
    def environment_name(self) -> Optional[str]:
        meta = self.meta or {}
        fal = meta.get("fal") or {}
        return fal.get("environment")


@dataclass
class DbtRunResult:
    native_run_result: Optional[RunResultsArtifact]

    @property
    @deprecated(details="Use native_run_result instead")
    def nativeRunResult(self):
        return self.native_run_result

    @property
    def results(self) -> Sequence[RunResultOutput]:
        if self.native_run_result:
            return self.native_run_result.results
        else:
            return []


@dataclass
class DbtFreshnessExecutionResult:
    _artifact: Optional[FreshnessExecutionResultArtifact]

    @property
    def results(self) -> Sequence[FreshnessNodeOutput]:
        if self._artifact:
            return self._artifact.results
        else:
            return []


@dataclass
class DbtManifest:
    native_manifest: Manifest

    @property
    @deprecated(details="Use native_manifest instead")
    def nativeManifest(self):
        return self.native_manifest

    def get_model_nodes(self) -> Iterable[ManifestNode]:
        return (
            node
            for node in self.native_manifest.nodes.values()
            if node.resource_type == NodeType.Model
        )

    def get_test_nodes(self) -> Iterable[ManifestNode]:
        return (
            node
            for node in self.native_manifest.nodes.values()
            if node.resource_type == NodeType.Test
        )

    def get_source_nodes(self) -> Iterable[SourceDefinition]:
        return self.native_manifest.sources.values()

    def _map_nodes(
        self,
        run_results: DbtRunResult,
        freshness_results: DbtFreshnessExecutionResult,
        generated_models: Dict[str, Path],
    ) -> Tuple[List[DbtModel], List[DbtSource], List[DbtTest]]:
        results_map = {r.unique_id: r for r in run_results.results}

        tests: List[DbtTest] = []

        tests_dict: Dict[str, List[DbtGenericTest]] = defaultdict(list)
        for node in self.get_test_nodes():
            test: DbtTest = DbtTest.init(node=node)

            result = results_map.get(node.unique_id)
            if result:
                test.status = result.status.value

            tests.append(test)

            if isinstance(test, DbtGenericTest):
                if test.model_id:
                    tests_dict[test.model_id].append(test)
                if test.source_id:
                    tests_dict[test.source_id].append(test)

        models: List[DbtModel] = []
        for node in self.get_model_nodes():
            model = DbtModel(
                node=node,
                tests=tests_dict[node.unique_id],
                python_model=generated_models.get(node.name),
            )

            result = results_map.get(node.unique_id)
            if result:
                model.status = result.status.value
                model.adapter_response = result.adapter_response

            models.append(model)

        source_freshness_map = {r.unique_id: r for r in freshness_results.results}

        sources: List[DbtSource] = []
        for node in self.get_source_nodes():
            source = DbtSource(
                node=node,
                tests=tests_dict[node.unique_id],
                freshness=source_freshness_map.get(node.unique_id),
            )

            result = results_map.get(node.unique_id)
            if result:
                source.status = result.status.value

            sources.append(source)

        return models, sources, tests


@dataclass
class CompileArgs:
    selector: Optional[str]
    select: List[str]
    models: List[str]
    exclude: Tuple[str]
    state: Optional[Path]
    single_threaded: Optional[bool]


@has_side_effects
class FalDbt:
    """Holds the entire dbt project information."""

    # TODO: figure out a meaningful __repr__ for this class
    def __init__(
        self,
        project_dir: Optional[str] = None,
        profiles_dir: Optional[str] = None,
        select: List[str] = [],
        exclude: Tuple[str] = tuple(),
        selector: Optional[str] = None,
        threads: Optional[int] = None,
        state: Optional[str] = None,
        profile_target: Optional[str] = None,
        args_vars: str = "{}",
        generated_models: Dict[str, Path] = {},
    ):
        if not version.is_version_plus("1.0.0"):
            raise NotImplementedError(
                f"dbt version {version.DBT_VCURRENT} is no longer supported, please upgrade to dbt 1.0.0 or above"
            )

        if project_dir is None:
            project_dir = os.getcwd()

        if profiles_dir is None:
            profiles_dir = str(default_profiles_dir())

        project_dir = os.path.realpath(os.path.expanduser(project_dir))
        profiles_dir = os.path.realpath(os.path.expanduser(profiles_dir))

        vars = parse.parse_cli_vars(args_vars)

        flags = lib.initialize_dbt_flags(
            profiles_dir=profiles_dir,
            project_dir=project_dir,
            threads=threads,
            profile_target=profile_target,
            vars=vars,
        )

        self.project_dir = flags.PROJECT_DIR
        self.profiles_dir = flags.PROFILES_DIR

        self._state = None
        if state is not None:
            self._state = Path(os.path.realpath(os.path.expanduser(state)))

        self.scripts_dir = parse.get_scripts_dir(self.project_dir, args_vars)


        # Can be overwritten if profile_target is not None
        self._config = parse.get_dbt_config(
            project_dir=self.project_dir,
            profiles_dir=self.profiles_dir,
            profile_target=profile_target,
            threads=threads,
            args_vars=args_vars,
        )

        self._run_results = DbtRunResult(
            parse.get_dbt_results(self.project_dir, self._config)
        )

        if self._run_results.native_run_result:
            if profile_target is None:
                profile_target = _get_custom_target(self._run_results)

        if profile_target is not None:
            self._config = parse.get_dbt_config(
                project_dir=self.project_dir,
                profiles_dir=self.profiles_dir,
                threads=threads,
                profile_target=profile_target,
                args_vars=args_vars,
            )

        lib.register_adapters(self._config)

        parse_result = self._dbt_invoke("parse")
        native_manifest: Manifest = parse_result.result # type: ignore

        # Necessary for manifest loading to not fail
        # dbt.tracking.initialize_tracking(self.profiles_dir)

        args = CompileArgs(selector, select, select, exclude, self._state, None)
        self._compile_task = CompileTask(args, self._config, native_manifest)

        self._compile_task._runtime_initialize()

        self._manifest = DbtManifest(native_manifest)

        freshness_execution_results = DbtFreshnessExecutionResult(
            parse.get_dbt_sources_artifact(self.project_dir, self._config)
        )

        self.models, self.sources, self.tests = self._manifest._map_nodes(
            self._run_results,
            freshness_execution_results,
            generated_models,
        )

        normalized_model_paths = parse.normalize_paths(
            self.project_dir, self.source_paths
        )

        self._global_script_paths = parse.get_global_script_configs(
            normalized_model_paths
        )

        self.features = self._find_features()
        self._environments = None

        telemetry.log_api(
            action="faldbt_initialized",
            dbt_config=self._config,
        )

    def _dbt_invoke(
        self, cmd: str, args: Optional[List[str]] = None
    ) -> dbtRunnerResult:
        runner = dbtRunner()

        if args is None:
            args = []

        project_args = [
            "--project-dir",
            self.project_dir,
            "--profiles-dir",
            self.profiles_dir,
            "--target",
            self._profile_target,
        ]

        # TODO: Intervene the dbt logs and capture them to avoid printing them to the console
        return runner.invoke([cmd] + project_args + args)

    @property
    def model_paths(self) -> List[str]:
        return self._config.model_paths

    @property
    @deprecated(details="Use model_paths instead")
    def source_paths(self) -> List[str]:
        return self.model_paths

    @property
    def _profile_target(self):
        return self._config.target_name

    @property
    def threads(self):
        return self._config.threads

    @property
    def target_path(self):
        return self._config.target_path

    @property
    def project_name(self):
        return self._config.project_name

    def list_sources(self) -> List[DbtSource]:
        """
        List tables available for `source` usage
        """
        with telemetry.log_time("list_sources", dbt_config=self._config):
            return self.sources

    def list_models_ids(self) -> Dict[str, str]:
        """
        List model ids available for `ref` usage, formatting like `[ref_name, ...]`
        """
        with telemetry.log_time("list_models_ids", dbt_config=self._config):
            res = {}
            for model in self.models:
                res[model.unique_id] = model.status

            return res

    def list_models(self) -> List[DbtModel]:
        """
        List models
        """
        with telemetry.log_time("list_models", dbt_config=self._config):
            return self.models

    def list_tests(self) -> List[DbtTest]:
        """
        List tests
        """
        with telemetry.log_time("list_tests", dbt_config=self._config):
            return self.tests

    def list_features(self) -> List[Feature]:
        with telemetry.log_time("list_features", dbt_config=self._config):
            return self.features

    def _find_features(self) -> List[Feature]:
        """List features defined in schema.yml files."""
        models = self.models
        models = list(
            filter(
                # Find models that have both feature store and column defs
                lambda model: FAL in model.meta
                and isinstance(model.meta[FAL], dict)
                and "feature_store" in model.meta[FAL]
                and len(list(model.columns.keys())) > 0,
                models,
            )
        )
        features = []
        for model in models:
            for column_name in model.columns.keys():
                if column_name == model.meta[FAL]["feature_store"]["entity_column"]:
                    continue
                if column_name == model.meta[FAL]["feature_store"]["timestamp_column"]:
                    continue
                features.append(
                    Feature(
                        model=model.name,
                        column=column_name,
                        description=model.columns[column_name].description,
                        entity_column=model.meta[FAL]["feature_store"]["entity_column"],
                        timestamp_column=model.meta[FAL]["feature_store"][
                            "timestamp_column"
                        ],
                    )
                )
        return features

    def _model(
        self, target_model_name: str, target_package_name: Optional[str]
    ) -> ManifestNode:
        # HACK: always setting node package as self.project_dir
        target_model: MaybeNonSource = self._manifest.native_manifest.resolve_ref(
            target_model_name,
            target_package_name,
            None,
            self.project_dir,
            self.project_dir,
        )
        package_str = f"'{target_package_name}'." if target_package_name else ""
        model_str = f"{package_str}'{target_model_name}'"
        if target_model is None:
            raise Exception(f"Could not find model {model_str}")

        if isinstance(target_model, Disabled):
            raise RuntimeError(f"Model {model_str} is disabled")

        return target_model

    def ref(self, target_1: str, target_2: Optional[str] = None) -> pd.DataFrame:
        """
        Download a dbt model as a pandas.DataFrame automagically.
        """
        with telemetry.log_time("ref", dbt_config=self._config):
            target_model_name = target_1
            target_package_name = None
            if target_2 is not None:
                target_package_name = target_1
                target_model_name = target_2

            target_model = self._model(target_model_name, target_package_name)

            return lib.fetch_target(
                self.project_dir,
                self.profiles_dir,
                target_model,
                self._profile_target,
                config=self._config,
            )

    def _source(
        self, target_source_name: str, target_table_name: str
    ) -> SourceDefinition:
        # HACK: always setting node package as self.project_dir
        target_source: MaybeParsedSource = (
            self._manifest.native_manifest.resolve_source(
                target_source_name,
                target_table_name,
                self.project_dir,
                self.project_dir,
            )
        )

        if target_source is None:
            raise RuntimeError(
                f"Could not find source '{target_source_name}'.'{target_table_name}'"
            )

        if isinstance(target_source, Disabled):
            raise RuntimeError(
                f"Source '{target_source_name}'.'{target_table_name}' is disabled"
            )

        return target_source

    def source(self, target_source_name: str, target_table_name: str) -> pd.DataFrame:
        """
        Download a dbt source as a pandas.DataFrame automagically.
        """
        with telemetry.log_time("source", dbt_config=self._config):
            target_source = self._source(target_source_name, target_table_name)

            return lib.fetch_target(
                self.project_dir,
                self.profiles_dir,
                target_source,
                self._profile_target,
                config=self._config,
            )

    def write_to_source(
        self,
        data: pd.DataFrame,
        target_source_name: str,
        target_table_name: str,
        *,
        dtype: Any = None,
        mode: str = "append",
    ):
        """
        Write a pandas.DataFrame to a dbt source automagically.
        """

        with telemetry.log_time(
            "write_to_source",
            dbt_config=self._config,
            additional_props={"args": {"mode": mode}},
        ):
            target_source = self._source(target_source_name, target_table_name)

            write_mode = lib.WriteModeEnum(mode.lower().strip())
            if write_mode == lib.WriteModeEnum.APPEND:
                lib.write_target(
                    data,
                    self.project_dir,
                    self.profiles_dir,
                    self._profile_target,
                    target_source,
                    dtype=dtype,
                    config=self._config,
                )

            elif write_mode == lib.WriteModeEnum.OVERWRITE:
                lib.overwrite_target(
                    data,
                    self.project_dir,
                    self.profiles_dir,
                    self._profile_target,
                    target_source,
                    dtype=dtype,
                    config=self._config,
                )

            else:
                raise Exception(f"write_to_source mode `{mode}` not supported")

    def write_to_model(
        self,
        data: pd.DataFrame,
        target_1: str,
        target_2: Optional[str] = None,
        *,
        dtype: Any = None,
        mode: str = "overwrite",
    ):
        """
        Write a pandas.DataFrame to a dbt model automagically.
        """

        with telemetry.log_time(
            "write_to_model",
            dbt_config=self._config,
            additional_props={"args": {"mode": mode}},
        ):
            target_model_name = target_1
            target_package_name = None
            if target_2 is not None:
                target_package_name = target_1
                target_model_name = target_2

            target_model = self._model(target_model_name, target_package_name)

            write_mode = lib.WriteModeEnum(mode.lower().strip())
            if write_mode == lib.WriteModeEnum.APPEND:
                lib.write_target(
                    data,
                    self.project_dir,
                    self.profiles_dir,
                    self._profile_target,
                    target_model,
                    dtype=dtype,
                    config=self._config,
                )

            elif write_mode == lib.WriteModeEnum.OVERWRITE:
                lib.overwrite_target(
                    data,
                    self.project_dir,
                    self.profiles_dir,
                    self._profile_target,
                    target_model,
                    dtype=dtype,
                    config=self._config,
                )

            else:
                raise Exception(f"write_to_model mode `{mode}` not supported")

    def execute_sql(self, sql: str) -> pd.DataFrame:
        """Execute a sql query."""

        with telemetry.log_time("execute_sql", dbt_config=self._config):
            # HACK: we need to pass config in because of weird behavior of execute_sql when
            # ran from GitHub Actions. For some reason, it can not find the right profile.
            # Haven't been able to reproduce this behavior locally and therefore developed
            # this workaround.
            compiled_result = lib.compile_sql(
                self.project_dir,
                self.profiles_dir,
                self._profile_target,
                sql,
                config=self._config,
            )

            # HACK: we need to pass config in because of weird behavior of execute_sql when
            # ran from GitHub Actions. For some reason, it can not find the right profile.
            # Haven't been able to reproduce this behavior locally and therefore developed
            # this workaround.

            # NOTE: changed in version 1.3.0 to `compiled_code`
            if hasattr(compiled_result, "compiled_code"):
                sql = compiled_result.compiled_code
            else:
                sql = compiled_result.compiled_sql
            return lib.execute_sql(
                self.project_dir,
                self.profiles_dir,
                self._profile_target,
                sql,
                config=self._config,
            )

    def _load_environment(self, name: str) -> "BaseEnvironment":
        """
        Return the environment for the given ``name``.
        If the environment does not exist, it raises an exception.
        """
        if self._environments is None:
            self._environments = parse.load_environments(self.project_dir)
        return self._environments[name]


def _get_custom_target(run_results: DbtRunResult):
    if "target" in run_results.native_run_result.args:
        return run_results.native_run_result.args["target"]
    return None



from dbt.adapters.fal_experimental.connections import TeleportCredentials
from dbt.fal.adapters.teleport.impl import TeleportAdapter
from dbt.fal.adapters.teleport.info import TeleportInfo
from dbt.adapters.base.relation import BaseRelation
from dbt.adapters.base.impl import BaseAdapter


class SnowflakeAdapterTeleport(TeleportAdapter):

    def __init__(self, db_adapter: BaseAdapter, teleport_credentials: TeleportCredentials):
        from dbt.adapters.fal_experimental.adapter_support import new_connection
        self._db_adapter = db_adapter
        self._credentials = teleport_credentials

        # TODO: put this in teleport_info
        url = f's3://{teleport_credentials.s3_bucket}/teleport'

        with new_connection(self._db_adapter, "fal-snowflake:setup-teleport") as conn:
            cur = conn.handle.cursor()

            create_stage_query = f"""CREATE OR REPLACE STAGE falstage
            URL = '{url}' CREDENTIALS = (
            aws_key_id='{self._credentials.s3_access_key_id}',
            aws_secret_key='{self._credentials.s3_access_key}');"""

            create_format_query = """CREATE OR REPLACE FILE FORMAT falparquet type = 'PARQUET';"""

            cur.execute(create_stage_query)
            cur.execute(create_format_query)


    @classmethod
    def storage_formats(cls):
        return ['parquet']

    def teleport_from_external_storage(self, relation: BaseRelation, relation_path: str, teleport_info: TeleportInfo) -> None:
        assert teleport_info.format == 'parquet', "snowflake only supports parquet format for Teleport"
        location = f"@falstage/{relation_path}"

        with self._db_adapter.connection_named('teleport:copy_from'):
            create_macro = self._db_adapter.execute_macro(
                'snowflake__create_table_from_parquet',
                kwargs={
                    'relation': relation,
                    'location': location,
                })

            self._db_adapter.execute(create_macro)

            copy_macro = self._db_adapter.execute_macro(
                'snowflake__copy_from_parquet',
                kwargs={
                    'relation': relation,
                    'location': location,
                }
            )
            self._db_adapter.execute(copy_macro)

    def teleport_to_external_storage(self, relation: BaseRelation, teleport_info: TeleportInfo) -> str:
        assert teleport_info.format == 'parquet', "snowflake only supports parquet format for Teleport"
        relation_path = teleport_info.build_relation_path(relation)
        location = f"@falstage/{relation_path}"
        rendered_macro = self._db_adapter.execute_macro(
            'snowflake__copy_to_parquet',
            kwargs={
                'relation': relation,
                'location': location
            })
        with self._db_adapter.connection_named('teleport:copy_to'):
            self._db_adapter.execute(rendered_macro)
        return relation_path




import argparse
from enum import Enum, auto
from concurrent.futures import (
    FIRST_COMPLETED,
    Executor,
    Future,
    ThreadPoolExecutor,
    wait,
)
from dataclasses import dataclass, field
from typing import Iterator, List, Optional

from fal.dbt.planner.schedule import SUCCESS, Scheduler
from fal.dbt.planner.tasks import (
    TaskGroup,
    Task,
    Status,
    DBTTask,
    FalLocalHookTask,
    HookType,
)
from fal.dbt.integration.project import FalDbt

from fal.dbt.integration.logger import LOGGER


class State(Enum):
    PRE_HOOKS = auto()
    MAIN_TASK = auto()
    POST_HOOKS = auto()


def _collect_nodes(groups: List[TaskGroup], fal_dbt: FalDbt) -> Iterator[str]:
    for group in groups:
        if isinstance(group.task, DBTTask):
            yield from group.task.model_ids
        if (
            isinstance(group.task, FalLocalHookTask)
            and group.task.hook_type is HookType.SCRIPT
        ):
            # Is a before/after script
            yield group.task.build_fal_script(fal_dbt).id


def _show_failed_groups(scheduler: Scheduler, fal_dbt: FalDbt) -> None:
    failed_nodes = list(
        _collect_nodes(scheduler.filter_groups(Status.FAILURE), fal_dbt)
    )
    if failed_nodes:
        message = ", ".join(failed_nodes)
        LOGGER.info("Failed calculating the following nodes: {}", message)

    skipped_nodes = list(
        _collect_nodes(scheduler.filter_groups(Status.SKIPPED), fal_dbt)
    )
    if skipped_nodes:
        message = ", ".join(skipped_nodes)
        LOGGER.info("Skipped calculating the following nodes: {}", message)


@dataclass
class FutureGroup:
    args: argparse.Namespace
    fal_dbt: FalDbt
    task_group: TaskGroup
    executor: Executor
    futures: List[Future] = field(default_factory=list)
    status: int = SUCCESS
    state: Optional[State] = None

    def __post_init__(self) -> None:
        if self.task_group.pre_hooks:
            self.switch_to(State.PRE_HOOKS)
        else:
            self.switch_to(State.MAIN_TASK)

    def switch_to(self, group_status: State) -> None:
        self.state = group_status
        if group_status is State.PRE_HOOKS:
            self._add_tasks(*self.task_group.pre_hooks)
        elif group_status is State.MAIN_TASK:
            self._add_tasks(self.task_group.task)
        else:
            assert group_status is State.POST_HOOKS
            self._add_tasks(*self.task_group.post_hooks)

    def process(self, future: Future) -> None:
        assert future in self.futures
        self.futures.remove(future)

        # If non-zero, it will remain non-zero
        self.status |= future.result()

        if self.futures:
            return None

        if self.state is State.PRE_HOOKS:
            # If there are no tasks left and the previous group was pre-hooks,
            # we'll decide based on the exit status (success => move to the main task,
            # failure => run the post-hooks and then exit).
            if self.status == SUCCESS:
                self.switch_to(State.MAIN_TASK)
            else:
                self.switch_to(State.POST_HOOKS)
        elif self.state is State.MAIN_TASK:
            # If we just executed the main task, then we'll proceed to the post-hooks
            self.switch_to(State.POST_HOOKS)
        else:
            # If we just executed post-hooks and there are no more tasks left,
            # we'll just exit ¯\_(ツ)_/¯
            assert self.state is State.POST_HOOKS
            return None

    def _add_tasks(self, *tasks: Task) -> None:
        for task in tasks:
            future = self.executor.submit(
                task.execute,
                args=self.args,
                fal_dbt=self.fal_dbt,
            )
            future.task, future.group = task, self
            self.futures.append(future)

    @property
    def is_done(self) -> int:
        return len(self.futures) == 0


def parallel_executor(
    args: argparse.Namespace,
    fal_dbt: FalDbt,
    scheduler: Scheduler,
) -> int:
    def get_futures(future_groups):
        return {
            # Unpack all running futures into a single set
            # to be consumed by wait().
            future
            for future_group in future_groups
            for future in future_group.futures
        }

    def create_futures(executor: ThreadPoolExecutor) -> List[FutureGroup]:
        return [
            # FutureGroup's are the secondary layer of the executor,
            # managing the parallelization of tasks.
            FutureGroup(
                args,
                fal_dbt,
                task_group=task_group,
                executor=executor,
            )
            for task_group in scheduler.iter_available_groups()
        ]

    with ThreadPoolExecutor(fal_dbt.threads) as executor:
        future_groups = create_futures(executor)
        futures = get_futures(future_groups)
        while futures:
            # Get the first completed futures, mark them done.
            completed_futures, _ = wait(futures, return_when=FIRST_COMPLETED)
            for future in completed_futures:
                group: FutureGroup = future.group
                group.process(future)
                if group.is_done:
                    scheduler.finish(group.task_group, status=group.status)

            # And load all the tasks that were blocked by those futures.
            future_groups.extend(create_futures(executor))
            futures = get_futures(future_groups)

    _show_failed_groups(scheduler, fal_dbt)

    return _exit_code(scheduler)


def _exit_code(scheduler: Scheduler) -> int:
    return int(any(scheduler.filter_groups(Status.FAILURE)))



import csv
from io import StringIO

import pandas as pd
import sqlalchemy
import awswrangler as wr

from dbt.adapters.base import BaseRelation
from dbt.adapters.base.connections import AdapterResponse

from dbt.adapters.fal_experimental.adapter_support import new_connection, drop_relation_if_it_exists

from dbt.adapters.redshift import RedshiftAdapter


def read_relation_as_df(
    adapter: RedshiftAdapter, relation: BaseRelation
) -> pd.DataFrame:
    sql = f"SELECT * FROM {relation}"

    assert adapter.type() == "redshift"

    with new_connection(adapter, "fal-redshift:read_relation_as_df") as conn:
        df = wr.redshift.read_sql_query(sql, con=conn.handle)
        return df


def write_df_to_relation(
    adapter: RedshiftAdapter,
    data: pd.DataFrame,
    relation: BaseRelation,
) -> AdapterResponse:

    assert adapter.type() == "redshift"

    with new_connection(adapter, "fal-redshift:write_df_to_relation") as connection:
        # TODO: this should probably live in the materialization macro.
        temp_relation = relation.replace_path(
            identifier=f"__dbt_fal_temp_{relation.identifier}"
        )
        drop_relation_if_it_exists(adapter, temp_relation)

        wr.redshift.to_sql(
            data,
            connection.handle,
            table=temp_relation.identifier,
            schema=temp_relation.schema,
            index=False,
        )

        adapter.cache.add(temp_relation)
        drop_relation_if_it_exists(adapter, relation)
        adapter.rename_relation(temp_relation, relation)
        adapter.commit_if_has_connection()

        return AdapterResponse("OK")



from typing import Optional, List, cast
from dbt.cli.main import dbtRunner, dbtRunnerResult

class falProject:
    """
    Represents a dbt project and access to its resources and utility functions.
    """

    def _dbt_invoke(
        self, cmd: str, args: Optional[List[str]] = None
    ) -> dbtRunnerResult:
        if args is None:
            args = []

        project_args = [
            "--project-dir",
            self.project_dir,
            "--profiles-dir",
            self.profiles_dir,
        ]
        if self.target_name:
            project_args.extend(["--target", self.target_name])

        # TODO: Intervene the dbt logs and capture them to avoid printing them to the console
        return self._runner.invoke([cmd] + project_args + args)

    def __init__(
        self, project_dir: str, profiles_dir: str, target_name: Optional[str] = None
    ):
        # TODO: Make project_dir and profiles_dir optional and use the current working directory and default profiles dir?
        self.project_dir = project_dir
        self.profiles_dir = profiles_dir
        self.target_name = target_name

        # Load the manifest information
        self._runner = dbtRunner()
        parse_result = self._dbt_invoke("parse")
        native_manifest = cast(Manifest, parse_result.result)  # type: ignore
        self._manifest = native_manifest
        # self._manifest = DbtManifest(native_manifest)

        # TODO: Do we need the manifest in there?
        self._runner = dbtRunner(manifest=native_manifest)



from __future__ import annotations

import os
import shutil
import subprocess
import sysconfig
import threading
from collections import defaultdict
from contextlib import ExitStack, contextmanager, nullcontext
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, ContextManager, Generic, Iterator, TypeVar, Dict, Any

from platformdirs import user_cache_dir

from fal.dbt.integration.logger import LOGGER
from fal.dbt.packages import bridge, isolated_runner

BASE_CACHE_DIR = Path(user_cache_dir("fal", "fal"))
BASE_CACHE_DIR.mkdir(exist_ok=True)

T = TypeVar("T")
K = TypeVar("K", bound="BaseEnvironment")
BasicCallable = Callable[[], int]


@contextmanager
def rmdir_on_fail(path: Path) -> Iterator[None]:
    try:
        yield
    except Exception:
        if path.exists():
            shutil.rmtree(path)
        raise


def log_env(env: BaseEnvironment, message: str, *args, kind: str = "trace", **kwargs):
    message = f"[{env.key}] {message}"
    log_method = getattr(LOGGER, kind)
    log_method(message, *args, **kwargs)


class BaseEnvironment(Generic[T]):
    def __init_subclass__(cls, make_thread_safe: bool = False) -> None:
        if make_thread_safe:
            lock_cls = threading.Lock
        else:
            lock_cls = nullcontext

        cls.lock_manager = defaultdict(lock_cls)
        return cls

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> BaseEnvironment:
        """Create a new environment from the given configuration."""
        raise NotImplementedError

    @property
    def key(self) -> str:
        """A unique identifier for this environment (combination of requirements,
        python version and other relevant information) that can be used for caching
        and identification purposes."""
        raise NotImplementedError

    def _get_or_create(self) -> T:
        """Implementation of the environment creation and retrieval behavior. Not
        thread safe."""
        raise NotImplementedError

    def get_or_create(self) -> T:
        """If the environment exists, return the connection info for it. If not,
        setup the environment and return it first and then return the newly constructed
        information. Thread safe."""
        with self.lock_manager[self.key]:
            return self._get_or_create()

    def open_connection(self, conn_info: T) -> EnvironmentConnection:
        """Return a new connection to the environment residing inside
        given path."""
        raise NotImplementedError

    @contextmanager
    def connect(self) -> Iterator[EnvironmentConnection]:
        env_info = self.get_or_create()
        with self.open_connection(env_info) as connection:
            yield connection


@dataclass
class EnvironmentConnection(Generic[K]):
    env: K

    def __enter__(self) -> EnvironmentConnection:
        return self

    def __exit__(self, *exc_info):
        return None

    def run(self, executable: BasicCallable, *args, **kwargs) -> Any:
        raise NotImplementedError


@dataclass
class IsolatedProcessConnection(EnvironmentConnection[K]):
    def run(self, executable: BasicCallable, *args, **kwargs) -> int:
        with ExitStack() as stack:
            # IPC flow is the following:
            #  1. [controller]: Create the socket server
            #  2. [controller]: Spawn the isolated process with the socket address
            #  3.   [isolated]: Connect to the socket server
            #  4. [controller]: Accept the incoming connection request
            #  5. [controller]: Send the executable over the established bridge
            #  6.   [isolated]: Receive the executable from the bridge
            #  7.   [isolated]: Execute the executable and once done send the result back
            #  8. [controller]: Loop until either the isolated process exits or sends any
            #                   data (will be interpreted as a tuple of two mutually exclusive
            #                   objects, either a result object or an exception to be raised).
            #

            log_env(self.env, "Starting the controller bridge.")
            controller_service = stack.enter_context(bridge.controller_connection())

            log_env(
                self.env,
                "Controller server is listening at {}.",
                controller_service.address,
            )
            isolated_process = stack.enter_context(
                self.start_process(controller_service, *args, **kwargs)
            )

            log_env(
                self.env,
                "Awaiting child process of {} to establish a connection.",
                isolated_process.pid,
            )
            established_connection = stack.enter_context(controller_service.accept())

            log_env(
                self.env,
                "Bridge between controller and the child has been established at {}.",
                controller_service.address,
            )
            established_connection.send(executable)

            log_env(
                self.env,
                "Executable has been sent, awaiting execution result and logs.",
            )
            return self.poll_until_result(isolated_process, established_connection)

    def start_process(
        self, connection: bridge.ConnectionWrapper, *args, **kwargs
    ) -> ContextManager[subprocess.Popen]:
        raise NotImplementedError

    def poll_until_result(
        self, process: subprocess.Popen, connection: bridge.ConnectionWrapper
    ) -> Any:
        while process.poll() is None:
            # Normally, if we do connection.read() without having this loop
            # it is going to block us indefinitely (even if the underlying
            # process has crashed). We can use a combination of process.poll
            # and connection.poll to check if the process is alive and has data
            # to move forward.
            if not connection.poll():
                continue

            try:
                result, exception = connection.recv()
            except EOFError:
                log_env(self.env, "The isolated process has unexpectedly exited.", kind="error")
                raise RuntimeError("The isolated process has unexpectedly exited.")

            if exception is None:
                log_env(
                    self.env, "Isolated process has returned the result: {}", result
                )
                return result
            else:
                log_env(
                    self.env,
                    "The isolated process has exited with an exception.",
                    kind="error",
                )
                raise exception


@dataclass
class DualPythonIPC(IsolatedProcessConnection[BaseEnvironment]):
    # We manage user-defined dual-Python environments in two steps.
    #   1. Create a primary environment which contains the default dependencies
    #      to run a fal script (like fal, dbt-core, and required dbt adapters).
    #   2. Create a secondary environment which contains the user-defined dependencies.
    #
    # This is an optimization we apply to reduce the cost of user-defined environments
    # where we can actually share the primary environment and save a lot of time from not
    # installing heavy dependencies like dbt adapters again and again. This also heavily
    # reduces the disk usage.

    primary_path: Path
    secondary_path: Path

    def start_process(
        self,
        service: bridge.Listener,
        *args,
        **kwargs,
    ) -> ContextManager[subprocess.Popen]:
        # We are going to use the primary environment to run the Python
        # interpreter, but at the same time we are going to inherit all
        # the packages from the secondary environment (user's environment)
        # so that they can technically override anything.

        # The search order is important, we want the secondary path to
        # take precedence.
        python_path = python_path_for(self.secondary_path, self.primary_path)

        # The environment which the python executable is going to be used
        # shouldn't matter much for the virtual-env based installations but
        # in conda, the Python executable between the primary and the secondary
        # environment might differ and we'll give precedence to the user's
        # choice (the secondary environment).
        python_executable = get_executable_path(self.secondary_path, "python")
        return subprocess.Popen(
            [
                python_executable,
                isolated_runner.__file__,
                bridge.encode_service_address(service.address),
            ],
            env={"PYTHONPATH": python_path, **os.environ},
        )


def python_path_for(*search_paths) -> str:
    assert len(search_paths) >= 1
    return os.pathsep.join(
        # sysconfig takes the virtual environment path and returns
        # the directory where all the site packages are located.
        sysconfig.get_path("purelib", vars={"base": search_path})
        for search_path in search_paths
    )


def get_executable_path(search_path: Path, executable_name: str) -> Path:
    bin_dir = (search_path / "bin").as_posix()
    executable_path = shutil.which(executable_name, path=bin_dir)
    if executable_path is None:
        raise RuntimeError(
            f"Could not find {executable_name} in {search_path}. "
            f"Is the virtual environment corrupted?"
        )

    return Path(executable_path)



import abc
import os
from time import sleep
import sys
from multiprocessing.synchronize import RLock
from threading import get_ident
from typing import (
    Any,
    Dict,
    Tuple,
    Hashable,
    Optional,
    List,
    Type,
    Union,
    Iterable,
    Callable,
)

from dbt.exceptions import (
    NotImplementedError,
    InvalidConnectionError,
    DbtInternalError,
    CompilationError,
    FailedToConnectError,
)
from dbt.contracts.connection import (
    Connection,
    Identifier,
    ConnectionState,
    AdapterRequiredConfig,
    LazyHandle,
    AdapterResponse,
)
from dbt.events import AdapterLogger
from dbt.events.functions import fire_event
from dbt.events.types import (
    NewConnection,
    ConnectionReused,
    ConnectionLeftOpen,
    ConnectionClosed,
)
from dbt import flags

SleepTime = Union[int, float]  # As taken by time.sleep.
AdapterHandle = Any  # Adapter connection handle objects can be any class.


class PythonConnectionManager(metaclass=abc.ABCMeta):
    """Methods to implement:
        - open
        - execute

    You must also set the 'TYPE' class attribute with a class-unique constant
    string.
    """

    TYPE: str = NotImplemented

    def __init__(self, profile: AdapterRequiredConfig):
        self.profile = profile
        if profile.credentials.type == self.TYPE:
            self.credentials = profile.credentials
        else:
            self.credentials = profile.python_adapter_credentials

        self.thread_connections: Dict[Hashable, Connection] = {}
        self.lock: RLock = flags.MP_CONTEXT.RLock()

    @staticmethod
    def get_thread_identifier() -> Hashable:
        # note that get_ident() may be re-used, but we should never experience
        # that within a single process
        return (os.getpid(), get_ident())

    def get_thread_connection(self) -> Connection:
        key = self.get_thread_identifier()
        with self.lock:
            if key not in self.thread_connections:
                raise InvalidConnectionError(key, list(self.thread_connections))
            return self.thread_connections[key]

    def set_thread_connection(self, conn: Connection) -> None:
        key = self.get_thread_identifier()
        if key in self.thread_connections:
            raise DbtInternalError(
                "In set_thread_connection, existing connection exists for {}"
            )
        self.thread_connections[key] = conn

    def get_if_exists(self) -> Optional[Connection]:
        key = self.get_thread_identifier()
        with self.lock:
            return self.thread_connections.get(key)

    def clear_thread_connection(self) -> None:
        key = self.get_thread_identifier()
        with self.lock:
            if key in self.thread_connections:
                del self.thread_connections[key]

    def set_connection_name(self, name: Optional[str] = None) -> Connection:
        conn_name: str
        if name is None:
            # if a name isn't specified, we'll re-use a single handle
            # named 'master'
            conn_name = "master"
        else:
            if not isinstance(name, str):
                raise CompilationError(
                    f"For connection name, got {name} - not a string!"
                )
            assert isinstance(name, str)
            conn_name = name

        conn = self.get_if_exists()
        if conn is None:
            conn = Connection(
                type=Identifier(self.TYPE),
                name=None,
                state=ConnectionState.INIT,
                transaction_open=False,
                handle=None,
                credentials=self.credentials,
            )
            self.set_thread_connection(conn)

        if conn.name == conn_name and conn.state == "open":
            return conn

        fire_event(NewConnection(conn_name=conn_name, conn_type=self.TYPE))

        if conn.state == "open":
            fire_event(ConnectionReused(conn_name=conn_name))
        else:
            conn.handle = LazyHandle(self.open)

        conn.name = conn_name
        return conn

    @classmethod
    def retry_connection(
        cls,
        connection: Connection,
        connect: Callable[[], AdapterHandle],
        logger: AdapterLogger,
        retryable_exceptions: Iterable[Type[Exception]],
        retry_limit: int = 1,
        retry_timeout: Union[Callable[[int], SleepTime], SleepTime] = 1,
        _attempts: int = 0,
    ) -> Connection:
        """Given a Connection, set its handle by calling connect.

        The calls to connect will be retried up to retry_limit times to deal with transient
        connection errors. By default, one retry will be attempted if retryable_exceptions is set.

        :param Connection connection: An instance of a Connection that needs a handle to be set,
            usually when attempting to open it.
        :param connect: A callable that returns the appropiate connection handle for a
            given adapter. This callable will be retried retry_limit times if a subclass of any
            Exception in retryable_exceptions is raised by connect.
        :type connect: Callable[[], AdapterHandle]
        :param AdapterLogger logger: A logger to emit messages on retry attempts or errors. When
            handling expected errors, we call debug, and call warning on unexpected errors or when
            all retry attempts have been exhausted.
        :param retryable_exceptions: An iterable of exception classes that if raised by
            connect should trigger a retry.
        :type retryable_exceptions: Iterable[Type[Exception]]
        :param int retry_limit: How many times to retry the call to connect. If this limit
            is exceeded before a successful call, a FailedToConnectError will be raised.
            Must be non-negative.
        :param retry_timeout: Time to wait between attempts to connect. Can also take a
            Callable that takes the number of attempts so far, beginning at 0, and returns an int
            or float to be passed to time.sleep.
        :type retry_timeout: Union[Callable[[int], SleepTime], SleepTime] = 1
        :param int _attempts: Parameter used to keep track of the number of attempts in calling the
            connect function across recursive calls. Passed as an argument to retry_timeout if it
            is a Callable. This parameter should not be set by the initial caller.
        :raises dbt.exceptions.FailedToConnectError: Upon exhausting all retry attempts without
            successfully acquiring a handle.
        :return: The given connection with its appropriate state and handle attributes set
            depending on whether we successfully acquired a handle or not.
        """
        timeout = retry_timeout(_attempts) if callable(retry_timeout) else retry_timeout
        if timeout < 0:
            raise FailedToConnectError(
                "retry_timeout cannot be negative or return a negative time."
            )

        if retry_limit < 0 or retry_limit > sys.getrecursionlimit():
            # This guard is not perfect others may add to the recursion limit (e.g. built-ins).
            connection.handle = None
            connection.state = ConnectionState.FAIL
            raise FailedToConnectError("retry_limit cannot be negative")

        try:
            connection.handle = connect()
            connection.state = ConnectionState.OPEN
            return connection

        except tuple(retryable_exceptions) as e:
            if retry_limit <= 0:
                connection.handle = None
                connection.state = ConnectionState.FAIL
                raise FailedToConnectError(str(e))

            logger.debug(
                f"Got a retryable error when attempting to open a {cls.TYPE} connection.\n"
                f"{retry_limit} attempts remaining. Retrying in {timeout} seconds.\n"
                f"Error:\n{e}"
            )

            sleep(timeout)
            return cls.retry_connection(
                connection=connection,
                connect=connect,
                logger=logger,
                retry_limit=retry_limit - 1,
                retry_timeout=retry_timeout,
                retryable_exceptions=retryable_exceptions,
                _attempts=_attempts + 1,
            )

        except Exception as e:
            connection.handle = None
            connection.state = ConnectionState.FAIL
            raise FailedToConnectError(str(e))

    @abc.abstractmethod
    def cancel(self, connection: Connection):
        """Cancel the given connection."""
        raise NotImplementedError("`cancel` is not implemented for this adapter!")

    def cancel_open(self) -> List[str]:
        names = []
        this_connection = self.get_if_exists()
        with self.lock:
            for connection in self.thread_connections.values():
                if connection is this_connection:
                    continue

                # if the connection failed, the handle will be None so we have
                # nothing to cancel.
                if (
                    connection.handle is not None
                    and connection.state == ConnectionState.OPEN
                ):
                    self.cancel(connection)
                if connection.name is not None:
                    names.append(connection.name)
        return names

    @classmethod
    @abc.abstractmethod
    def open(cls, connection: Connection) -> Connection:
        """Open the given connection on the adapter and return it.

        This may mutate the given connection (in particular, its state and its
        handle).

        This should be thread-safe, or hold the lock if necessary. The given
        connection should not be in either in_use or available.
        """
        raise NotImplementedError("`open` is not implemented for this adapter!")

    def release(self) -> None:
        with self.lock:
            conn = self.get_if_exists()
            if conn is None:
                return

        try:
            # always close the connection
            self.close(conn)
        except Exception:
            # if rollback or close failed, remove our busted connection
            self.clear_thread_connection()
            raise

    def cleanup_all(self) -> None:
        with self.lock:
            for connection in self.thread_connections.values():
                if connection.state not in {"closed", "init"}:
                    fire_event(ConnectionLeftOpen(conn_name=connection.name))
                else:
                    fire_event(ConnectionClosed(conn_name=connection.name))
                self.close(connection)

            # garbage collect these connections
            self.thread_connections.clear()

    @classmethod
    def _close_handle(cls, connection: Connection) -> None:
        """Perform the actual close operation."""
        # On windows, sometimes connection handles don't have a close() attr.
        if hasattr(connection.handle, "close"):
            connection.handle.close()

    @classmethod
    def close(cls, connection: Connection) -> Connection:
        # if the connection is in closed or init, there's nothing to do
        if connection.state in {ConnectionState.CLOSED, ConnectionState.INIT}:
            return connection

        cls._close_handle(connection)
        connection.state = ConnectionState.CLOSED

        return connection

    @abc.abstractmethod
    def execute(self, compiled_code: str) -> Tuple[AdapterResponse, Any]:
        """Execute the given Python code.

        :param str compiled_code: The Python code to execute.
        :return: A tuple of the run status and results
                (type is left to be decided by the Adapter implementation for now).
        :rtype: Tuple[AdapterResponse, Any]
        """
        raise NotImplementedError("`execute` is not implemented for this adapter!")



from __future__ import annotations

from typing import Callable, Iterator, List, Set

import networkx as nx
from fal.dbt.node_graph import NodeKind
from fal.dbt.cli.selectors import ExecutionPlan, _is_before_script, _is_after_script
from fal.dbt.integration.logger import LOGGER
from dataclasses import dataclass


@dataclass
class OriginGraph:
    graph: nx.DiGraph

    def copy_graph(self) -> nx.DiGraph:
        return self.graph.copy()  # type: ignore

    def _plot(self, graph=None):
        """
        For development and debugging purposes
        """
        if not graph:
            graph = self.graph

        import matplotlib.pyplot as plt

        import networkx.drawing.layout as layout

        nx.draw_networkx(
            graph,
            arrows=True,
            pos=layout.circular_layout(graph),
            labels={
                node: node.replace(".", "\n")
                .replace("model\n", "")
                .replace("script\n", "")
                .replace("\npy", ".py")
                for node in graph.nodes
            },
        )
        plt.show()


@dataclass
class FilteredGraph(OriginGraph):
    graph: nx.DiGraph

    @classmethod
    def from_execution_plan(
        cls,
        origin_graph: OriginGraph,
        execution_plan: ExecutionPlan,
    ) -> FilteredGraph:
        graph = origin_graph.copy_graph()
        for node in origin_graph.graph.nodes:
            if node not in execution_plan.nodes:
                graph.remove_node(node)

        if execution_plan.after_scripts or execution_plan.before_scripts:
            LOGGER.warn(
                "Using before/after scripts are now deprecated. "
                "Please consider migrating to pre-hooks/post-hooks or Python models."
            )

        return cls(graph)


@dataclass
class ScriptConnectedGraph(OriginGraph):
    graph: nx.DiGraph

    @classmethod
    def from_filtered_graph(
        cls,
        filtered_graph: FilteredGraph,
    ):
        graph = filtered_graph.copy_graph()
        shuffled_graph = cls(graph)
        shuffled_graph._shuffle()
        return shuffled_graph

    def _shuffle(self):
        def _pattern_matching(node_check: Callable[[str], bool], nodes: Set[str]):
            matched = {
                maybe_script for maybe_script in nodes if node_check(maybe_script)
            }
            return matched, nodes - matched

        def get_before_scripts(graph: nx.DiGraph, node: str):
            return _pattern_matching(_is_before_script, set(graph.predecessors(node)))

        def get_after_scripts(graph: nx.DiGraph, node: str):
            return _pattern_matching(_is_after_script, set(graph.successors(node)))

        def _add_edges_from_to(from_nodes: Set[str], to_nodes: Set[str]):
            self.graph.add_edges_from(
                (from_n, to_n) for from_n in from_nodes for to_n in to_nodes
            )

        old_graph = self.copy_graph()
        node: str
        for node in old_graph.nodes:

            after_scripts, other_succs = get_after_scripts(old_graph, node)
            # Keep the original node to succs edges and add a new one from the script to succs
            _add_edges_from_to(after_scripts, other_succs)

            before_scripts, other_preds = get_before_scripts(old_graph, node)
            # Keep the original preds to node edge and add a new one from preds to the scripts
            _add_edges_from_to(other_preds, before_scripts)

            # Add edges between node's after and succ's before scripts
            for succ in other_succs:
                succ_before_scripts, _succ_other_preds = get_before_scripts(
                    old_graph, succ
                )

                # Add edge between all after scripts to the succ's before scripts
                _add_edges_from_to(after_scripts, succ_before_scripts)


@dataclass
class PlannedGraph(OriginGraph):
    graph: nx.DiGraph

    @classmethod
    def from_script_connected_graph(
        cls,
        shuffled_graph: ScriptConnectedGraph,
        enable_chunking: bool = True,
    ):
        graph = shuffled_graph.copy_graph()
        planned_graph = cls(graph)
        if enable_chunking:
            planned_graph.plan()
        return planned_graph

    def _find_subgraphs(self) -> Iterator[List[str]]:
        # Initially topologically sort the graph and find nodes
        # that can be grouped together to be run as a single node
        # by using the critical node approach. All nodes within a
        # group must have the same ancestors, to avoid removing
        # existing branches.

        current_stack = []
        allowed_ancestors = set()

        def split() -> Iterator[List[str]]:
            if len(current_stack) > 1:
                yield current_stack.copy()
            current_stack.clear()
            allowed_ancestors.clear()

        for node in nx.topological_sort(self.graph):
            properties = self.graph.nodes[node]
            if properties["kind"] in (
                NodeKind.FAL_MODEL,
                NodeKind.FAL_SCRIPT,
            ) or properties.get("pre_hook"):
                yield from split()
                continue

            ancestors = nx.ancestors(self.graph, node)
            if not current_stack:
                allowed_ancestors = ancestors

            if not ancestors.issubset(allowed_ancestors):
                yield from split()

            current_stack.append(node)
            allowed_ancestors |= {node, *ancestors}

            if properties.get("post_hook"):
                yield from split()

        yield from split()

    def _reduce_subgraph(
        self,
        nodes: List[str],
    ) -> None:
        subgraph = self.graph.subgraph(nodes).copy()

        # Use the same set of properties as the last
        # node, since only it can have any post hooks.
        self.graph.add_node(
            subgraph,
            **self.graph.nodes[nodes[-1]].copy(),
            exit_node=nodes[-1],
        )

        for node in nodes:
            for predecessor in self.graph.predecessors(node):
                self.graph.add_edge(predecessor, subgraph)

            for successor in self.graph.successors(node):
                self.graph.add_edge(subgraph, successor)

            self.graph.remove_node(node)

    def plan(self):
        # Implementation of Gorkem's Critical Nodes Algorithm
        # with a few modifications.
        for nodes in self._find_subgraphs():
            self._reduce_subgraph(nodes)



"""Fal utilities."""
import copy
from fal.dbt.integration.logger import LOGGER
from typing import List, TYPE_CHECKING, TypeVar

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache

if TYPE_CHECKING:
    from fal.dbt.fal_script import FalScript


def print_run_info(scripts: List["FalScript"]):
    """Print information on the current fal run."""
    models_str = "\n".join(map(lambda script: script.id, scripts))
    LOGGER.info(f"Starting fal run for following models and scripts: \n{models_str}\n")


class DynamicIndexProvider:
    def __init__(self) -> None:
        self._index = 0

    def next(self) -> int:
        """Increment the counter and return the last value."""
        index = self._index
        self._index += 1
        return index

    def __int__(self) -> int:
        """Return the last value."""
        return self._index


class _ReInitialize:
    def __init__(self, *args, **kwargs):
        self._serialization_state = {
            "args": copy.deepcopy(args),
            "kwargs": copy.deepcopy(kwargs),
        }
        super().__init__(*args, **kwargs)

    def __getstate__(self):
        return self._serialization_state

    def __setstate__(self, state):
        super().__init__(*state["args"], **state["kwargs"])

Cls = TypeVar("Cls", bound=type)

def has_side_effects(cls: "Cls") -> "Cls":
    """The given class has possible side-effects that might
    make the regular serialization problematic (e.g. registering
    adapters to DBT's global factory)."""
    return type(cls.__name__, (_ReInitialize, cls), {})


def cache_static(func):
    """Cache the result of a function."""
    return lru_cache(maxsize=None)(func)



from __future__ import annotations

import sys
import hashlib
import subprocess
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Dict, Any

from fal.dbt.packages.dependency_analysis import get_default_pip_dependencies
from fal.dbt.packages.environments.base import (
    BASE_CACHE_DIR,
    BaseEnvironment,
    DualPythonIPC,
    get_executable_path,
    log_env,
    python_path_for,
    rmdir_on_fail,
)
from fal.dbt.utils import cache_static

_BASE_VENV_DIR = BASE_CACHE_DIR / "venvs"
_BASE_VENV_DIR.mkdir(exist_ok=True)


@dataclass
class VirtualPythonEnvironment(BaseEnvironment[Path], make_thread_safe=True):
    requirements: List[str]
    inherit_from_local: bool = False

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> VirtualPythonEnvironment:
        requirements = config.get("requirements", [])
        inherit_from_local = config.get("_inherit_from_local", False)
        return cls(requirements, inherit_from_local=inherit_from_local)

    @property
    def key(self) -> str:
        return hashlib.sha256(" ".join(self.requirements).encode()).hexdigest()

    def _verify_dependencies(self, primary_path: Path, secondary_path: Path) -> None:
        # Ensure that there are no dependency mismatches between the
        # primary environment and the secondary environment.
        python_path = python_path_for(secondary_path, primary_path)
        original_pip = get_executable_path(primary_path, "pip")
        subprocess.check_call([original_pip, "check"], env={"PYTHONPATH": python_path})

    def _get_or_create(self) -> Path:
        from virtualenv import cli_run

        path = _BASE_VENV_DIR / self.key
        if path.exists():
            return path

        with rmdir_on_fail(path):
            log_env(self, "Creating virtual environment at {}", path, kind="info")
            cli_run([str(path)])
            log_env(
                self,
                "Installing requirements: {}",
                ", ".join(self.requirements),
                kind="info",
            )
            if self.requirements:
                pip_path = get_executable_path(path, "pip")
                subprocess.check_call([pip_path, "install"] + self.requirements)

            if not self.inherit_from_local:
                primary_env = get_primary_virtual_env()
                if self is not primary_env:
                    self._verify_dependencies(primary_env._get_or_create(), path)

        return path

    def open_connection(self, conn_info: Path) -> DualPythonIPC:
        if self.inherit_from_local:
            # Instead of creating a separate environment that only has
            # the same versions of fal/dbt-core etc. you have locally,
            # we can also use your environment as the primary. This is
            # mainly for the development time where the fal or dbt-core
            # you are using is not available on PyPI yet.
            primary_venv_path = Path(sys.exec_prefix)
        else:
            primary_venv = get_primary_virtual_env()
            primary_venv_path = primary_venv.get_or_create()
        secondary_venv_path = conn_info
        return DualPythonIPC(self, primary_venv_path, secondary_venv_path)


@cache_static
def get_primary_virtual_env() -> VirtualPythonEnvironment:
    return VirtualPythonEnvironment(get_default_pip_dependencies())



from __future__ import annotations

import zipfile
import io
from functools import partial
from tempfile import NamedTemporaryFile
from typing import Any, Optional

from dbt.adapters.base.impl import BaseAdapter
from dbt.config.runtime import RuntimeConfig
from dbt.contracts.connection import AdapterResponse
from dbt.flags import get_flags, Namespace

from fal_serverless import FalServerlessHost, isolated
from dbt.adapters.fal_experimental.utils.environments import (
    EnvironmentDefinition,
    get_default_pip_dependencies
)

from dbt.parser.manifest import MacroManifest, Manifest

from .adapter_support import (
    prepare_for_adapter,
    read_relation_as_df,
    reconstruct_adapter,
    write_df_to_relation,
)

from .utils import extra_path, get_fal_scripts_path, retrieve_symbol


def run_with_adapter(code: str, adapter: BaseAdapter, config: RuntimeConfig) -> Any:
    # main symbol is defined during dbt-fal's compilation
    # and acts as an entrypoint for us to run the model.
    fal_scripts_path = str(get_fal_scripts_path(config))
    with extra_path(fal_scripts_path):
        main = retrieve_symbol(code, "main")
        return main(
            read_df=prepare_for_adapter(adapter, read_relation_as_df),
            write_df=prepare_for_adapter(adapter, write_df_to_relation),
        )


def _isolated_runner(
    code: str,
    flags: Namespace,
    config: RuntimeConfig,
    manifest: Manifest,
    macro_manifest: MacroManifest,
    local_packages: Optional[bytes] = None,
) -> Any:
    # This function can be run in an entirely separate
    # process or an environment, so we need to reconstruct
    # the DB adapter solely from the config.
    adapter = reconstruct_adapter(flags, config, manifest, macro_manifest)
    fal_scripts_path = get_fal_scripts_path(config)
    if local_packages is not None:
        if fal_scripts_path.exists():
            import shutil
            shutil.rmtree(fal_scripts_path)
        fal_scripts_path.parent.mkdir(parents=True, exist_ok=True)
        zip_file = zipfile.ZipFile(io.BytesIO(local_packages))
        zip_file.extractall(fal_scripts_path)

    return run_with_adapter(code, adapter, config)


def run_in_environment_with_adapter(
    environment: EnvironmentDefinition,
    code: str,
    config: RuntimeConfig,
    manifest: Manifest,
    macro_manifest: MacroManifest,
    adapter_type: str
) -> AdapterResponse:
    """Run the 'main' function inside the given code on the
    specified environment.

    The environment_name must be defined inside fal_project.yml file
    in your project's root directory."""
    compressed_local_packages = None

    is_remote = type(environment.host) is FalServerlessHost

    deps = get_default_pip_dependencies(
        is_remote=is_remote,
        adapter_type=adapter_type)

    fal_scripts_path = get_fal_scripts_path(config)

    if is_remote and fal_scripts_path.exists():
        with NamedTemporaryFile() as temp_file:
            with zipfile.ZipFile(
                temp_file.name, "w", zipfile.ZIP_DEFLATED
            ) as zip_file:
                for entry in fal_scripts_path.rglob("*"):
                    zip_file.write(entry, entry.relative_to(fal_scripts_path))

            compressed_local_packages = temp_file.read()

    execute_model = partial(
        _isolated_runner,
        code=code,
        flags=get_flags(),
        config=config,
        manifest=manifest,
        macro_manifest=macro_manifest,
        local_packages=compressed_local_packages
    )

    if environment.kind == "local":
        result = execute_model()
        return result

    extra = {}
    if is_remote:
        # Machine type is only applicable in FalServerlesshost
        extra = {"machine_type": environment.machine_type}

    if environment.kind == "virtualenv":
        requirements = environment.config.get("requirements", [])
        requirements += deps
        isolated_function = isolated(
            kind="virtualenv",
            host=environment.host,
            requirements=requirements,
            **extra
        )(execute_model)
    elif environment.kind == "conda":
        dependencies = environment.config.pop("packages", [])
        dependencies.append({"pip": deps})
        env_dict = {
            "name": "dbt_fal_env",
            "channels": ["conda-forge", "defaults"],
            "dependencies": dependencies
        }
        isolated_function = isolated(
            kind="conda",
            host=environment.host,
            env_dict=env_dict,
            **extra
        )(execute_model)
    else:
        # We should not reach this point, because environment types are validated when the
        # environment objects are created (in utils/environments.py).
        raise Exception(f"Environment type not supported: {environment.kind}")

    result = isolated_function()
    return result



# NOTE: INSPIRED IN https://github.com/dbt-labs/dbt-core/blob/43edc887f97e359b02b6317a9f91898d3d66652b/core/dbt/lib.py
from contextlib import contextmanager
import six
from enum import Enum
from dataclasses import dataclass
from uuid import uuid4
from typing import Dict, Iterator, List, Optional, Tuple
from urllib.parse import quote_plus
import threading

import dbt.flags as flags
import dbt.adapters.factory as adapters_factory

from dbt.contracts.connection import AdapterResponse
from dbt.adapters.sql import SQLAdapter
from dbt.adapters.base import BaseRelation, BaseAdapter, BaseConnectionManager
from dbt.config import RuntimeConfig

from dbt.contracts.graph.nodes import ResultNode

import pandas as pd
from pandas.io import sql as pdsql

import agate
import sqlalchemy
from sqlalchemy.sql.ddl import CreateTable
from sqlalchemy.sql import Insert

from dbt.contracts.sql import RemoteRunResult


from fal.dbt.integration import parse
from fal.dbt.integration.logger import LOGGER


class WriteModeEnum(Enum):
    APPEND = "append"
    OVERWRITE = "overwrite"


@dataclass
class FlagsArgs:
    project_dir: str
    profiles_dir: str
    threads: Optional[int]
    profile: Optional[str]
    target: Optional[str]
    vars: Optional[Dict[str, str]]
    use_colors: Optional[bool]


def initialize_dbt_flags(
    profiles_dir: str,
    project_dir: str,
    profile_target: Optional[str],
    vars: Optional[dict],
    threads: Optional[int],
):
    """
    Initializes the flags module from dbt, since it's accessed from around their code.
    """
    args = FlagsArgs(
        use_colors=None,
        project_dir=project_dir,
        profiles_dir=profiles_dir,
        # TODO: accept profile flag too
        profile=None,
        target=profile_target,
        vars=vars,
        threads=threads,
    )

    flags.set_from_args(args, None)

    # Set invocation id
    import dbt.events.functions as events_functions

    events_functions.set_invocation_id()

    return flags.get_flags()


def register_adapters(config: RuntimeConfig):
    # HACK: to avoid 'Node package named <profile> not found'
    adapters_factory.reset_adapters()
    adapters_factory.register_adapter(config)


# NOTE: Once we get an adapter, we must call `connection_for` or `connection_named` to use it
def _get_adapter(
    project_dir: str,
    profiles_dir: str,
    profile_target: str,
    *,
    config: Optional[RuntimeConfig] = None,
) -> SQLAdapter:
    if config is None:
        config = parse.get_dbt_config(
            project_dir=project_dir,
            profiles_dir=profiles_dir,
            profile_target=profile_target,
        )
    adapter: SQLAdapter = adapters_factory.get_adapter(config)  # type: ignore

    return adapter


# HACK: to avoid https://github.com/uqfoundation/dill/issues/321
# When the lock is unpickled in the other thread, it is initialized as locked
if hasattr(threading, "_PyRLock"):
    _lock: threading.RLock = threading._PyRLock()  # type: ignore
else:
    _lock = threading.RLock()

@contextmanager
def _cache_lock(info: str = ""):
    global _lock

    operationId = uuid4()
    LOGGER.debug("Locking  {} {}", operationId, info)

    with _lock:
        yield


def _connection_name(prefix: str, obj, _hash: bool = True):
    # HACK: we need to include uniqueness (UUID4) to avoid clashes
    return f"{prefix}:{hash(str(obj)) if _hash else obj}:{uuid4()}"


def _execute_sql(
    adapter: SQLAdapter,
    sql: str,
    *,
    new_conn=True,
) -> Tuple[AdapterResponse, pd.DataFrame]:
    if adapter.type() == "bigquery":
        return _bigquery_execute_sql(adapter, sql, new_conn)

    if adapter.type() == "snowflake":
        return _snowflake_execute_sql(adapter, sql, new_conn)

    with _existing_or_new_connection(
        adapter, _connection_name("execute_sql", sql), new_conn
    ) as is_new:
        exec_response: Tuple[AdapterResponse, agate.Table] = adapter.execute(
            sql, auto_begin=is_new, fetch=True
        )
        response, agate_table = exec_response

        if is_new:
            adapter.commit_if_has_connection()

    return response, _agate_table_to_df(agate_table)


def _clear_relations_cache(adapter: BaseAdapter):
    # HACK: Sometimes we cache an incomplete cache or create stuff without the cache noticing.
    #       Some adapters work without this. We should separate adapter solutions like dbt.
    manifest = parse.get_dbt_manifest(adapter.config)
    adapter.set_relations_cache(manifest, True)


def _get_target_relation(
    adapter: SQLAdapter, target: ResultNode
) -> Optional[BaseRelation]:
    with adapter.connection_named(_connection_name("relation", target)):
        with _cache_lock("_get_target_relation"):
            _clear_relations_cache(adapter)

            # This ROLLBACKs so it has to be a new connection
            return adapter.get_relation(
                target.database, target.schema, target.identifier
            )


def compile_sql(
    project_dir: str,
    profiles_dir: str,
    profile_target: str,
    sql: str,
    *,
    config: Optional[RuntimeConfig] = None,
    adapter: Optional[SQLAdapter] = None,
):
    from dbt.parser.manifest import process_node
    from dbt.task.sql import SqlCompileRunner
    from dbt.parser.sql import SqlBlockParser

    if config is None:
        config = parse.get_dbt_config(
            project_dir=project_dir,
            profiles_dir=profiles_dir,
            profile_target=profile_target,
        )

    if adapter is None:
        adapter = _get_adapter(project_dir, profiles_dir, profile_target, config=config)

    manifest = parse.get_dbt_manifest(config)

    block_parser = SqlBlockParser(
        project=config,
        manifest=manifest,
        root_project=config,
    )

    sql_node = block_parser.parse_remote(sql, _connection_name("compile_sql", sql))
    process_node(config, manifest, sql_node)
    runner = SqlCompileRunner(config, adapter, sql_node, 1, 1)
    result: RemoteRunResult = runner.safe_run(manifest)
    return result


def execute_sql(
    project_dir: str,
    profiles_dir: str,
    profile_target: str,
    sql: str,
    *,
    config: Optional[RuntimeConfig] = None,
    adapter: Optional[SQLAdapter] = None,
) -> pd.DataFrame:
    if adapter is None:
        adapter = _get_adapter(project_dir, profiles_dir, profile_target, config=config)

    return _execute_sql(adapter, sql)[1]


def _agate_table_to_df(table: agate.Table) -> pd.DataFrame:
    column_names = list(table.column_names)
    rows = [list(row) for row in table]

    # TODO: better type matching?
    return pd.DataFrame.from_records(data=rows, columns=column_names, coerce_float=True)


def fetch_target(
    project_dir: str,
    profiles_dir: str,
    target: ResultNode,
    profile_target: str,
    *,
    config: Optional[RuntimeConfig] = None,
    adapter: Optional[SQLAdapter] = None,
) -> pd.DataFrame:
    if adapter is None:
        adapter = _get_adapter(project_dir, profiles_dir, profile_target, config=config)

    relation = _get_target_relation(adapter, target)

    if relation is None:
        raise Exception(f"Could not get relation for '{target.unique_id}'")

    return _fetch_relation(adapter, relation)


def _fetch_relation(adapter: SQLAdapter, relation: BaseRelation) -> pd.DataFrame:
    if adapter.type() == "postgres":
        return _sqlalchemy_engine_fetch_relation(adapter, relation)

    query = f"SELECT * FROM {relation}"
    return _execute_sql(adapter, query)[1]


def _build_table_from_parts(
    adapter: SQLAdapter,
    database: Optional[str],
    schema: Optional[str],
    identifier: Optional[str],
) -> BaseRelation:
    from dbt.contracts.relation import Path, RelationType

    path = Path(database, schema, identifier)

    # NOTE: assuming we want TABLE relation if not found
    return adapter.Relation(path, type=RelationType.Table)


def _build_table_from_target(adapter: SQLAdapter, target: ResultNode):
    return _build_table_from_parts(
        adapter, target.database, target.schema, target.identifier
    )


def overwrite_target(
    data: pd.DataFrame,
    project_dir: str,
    profiles_dir: str,
    profile_target: str,
    target: ResultNode,
    *,
    dtype=None,
    config: Optional[RuntimeConfig] = None,
    adapter: Optional[SQLAdapter] = None,
) -> AdapterResponse:
    if not adapter:
        adapter = _get_adapter(project_dir, profiles_dir, profile_target, config=config)

    relation = _build_table_from_target(adapter, target)

    if adapter.type() == "bigquery":
        return _bigquery_write_relation(
            adapter,
            data,
            relation,
            mode=WriteModeEnum.OVERWRITE,
            fields_schema=dtype,
        )

    # With some writing functions, it could be called twice at the same time for the same identifier
    # so we avoid overwriting temporal tables by attaching uniqueness to the name
    unique_str = str(uuid4())[0:8]
    temporal_relation = _build_table_from_parts(
        adapter,
        relation.database,
        relation.schema,
        f"{relation.identifier}__f__{unique_str}",
    )

    results = _write_relation(adapter, data, temporal_relation, dtype=dtype)

    try:
        _replace_relation(adapter, relation, temporal_relation)

        return results
    except:
        _drop_relation(adapter, temporal_relation)
        raise


def write_target(
    data: pd.DataFrame,
    project_dir: str,
    profiles_dir: str,
    profile_target: str,
    target: ResultNode,
    *,
    dtype=None,
    config: Optional[RuntimeConfig] = None,
    adapter: Optional[SQLAdapter] = None,
) -> AdapterResponse:
    if adapter is None:
        adapter = _get_adapter(project_dir, profiles_dir, profile_target, config=config)

    relation = _build_table_from_target(adapter, target)

    return _write_relation(adapter, data, relation, dtype=dtype)


def _write_relation(
    adapter: SQLAdapter,
    data: pd.DataFrame,
    relation: BaseRelation,
    *,
    dtype=None,
) -> AdapterResponse:
    if adapter.type() == "fal":
        adapter = adapter._db_adapter

    if adapter.type() == "bigquery":
        return _bigquery_write_relation(
            adapter,
            data,
            relation,
            mode=WriteModeEnum.APPEND,
            fields_schema=dtype,
        )

    if adapter.type() == "snowflake":
        return _snowflake_write_relation(
            adapter,
            data,
            relation,
        )

    if adapter.type() == "postgres":
        return _sqlalchemy_engine_write_relation(adapter, data, relation, dtype=dtype)

    database, schema, identifier = (
        relation.database,
        relation.schema,
        relation.identifier,
    )

    engine = _alchemy_mock_engine(adapter)
    pddb = pdsql.SQLDatabase(engine, schema=schema)
    pdtable = pdsql.SQLTable(identifier, pddb, data, index=False, dtype=dtype)

    alchemy_table: sqlalchemy.Table = pdtable.table.to_metadata(pdtable.pd_sql.meta)

    # HACK: athena needs "location" property that is not passed by mock adapter
    if adapter.type() == "athena":
        s3_dir = adapter.config.credentials.s3_staging_dir
        alchemy_table.dialect_options["awsathena"] = {
            "location": f"{s3_dir}{alchemy_table.schema}/{alchemy_table.name}/",
            "tblproperties": None,
            "compression": None,
            "bucket_count": None,
            "row_format": None,
            "serdeproperties": None,
            "file_format": None,
            "partition": None,
            "cluster": None,
        }

    column_names: List[str] = list(data.columns)

    rows = data.to_records(index=False)
    row_dicts = list(map(lambda row: dict(zip(column_names, row)), rows))

    create_stmt = CreateTable(alchemy_table, if_not_exists=True).compile(
        bind=engine, compile_kwargs={"literal_binds": True}
    )
    _execute_sql(adapter, six.text_type(create_stmt).strip())

    insert_stmt = Insert(alchemy_table, values=row_dicts).compile(
        bind=engine, compile_kwargs={"literal_binds": True}
    )
    response, _ = _execute_sql(adapter, six.text_type(insert_stmt).strip())
    return response


def _replace_relation(
    adapter: SQLAdapter,
    original_relation: BaseRelation,
    new_relation: BaseRelation,
):
    with adapter.connection_named(
        _connection_name("replace_relation", original_relation, _hash=False)
    ):
        with _cache_lock("_replace_relation"):
            adapter.connections.begin()

            _clear_relations_cache(adapter)

            if adapter.type() not in ("bigquery", "snowflake"):
                # This is a 'DROP ... IF EXISTS', so it always works
                adapter.drop_relation(original_relation)

            if adapter.type() == "athena":
                # HACK: athena doesn't support renaming tables, we do it manually
                create_stmt = f"create table {original_relation} as select * from {new_relation} with data"
                _execute_sql(
                    adapter,
                    six.text_type(create_stmt).strip(),
                    new_conn=False,
                )
                adapter.drop_relation(new_relation)
            elif adapter.type() == "bigquery":
                create_stmt = f"create or replace table {original_relation} as select * from {new_relation}"
                _bigquery_execute_sql(
                    adapter,
                    six.text_type(create_stmt).strip(),
                    new_conn=False,
                )
                adapter.drop_relation(new_relation)
            elif adapter.type() == "snowflake":
                create_stmt = (
                    f"create or replace table {original_relation} clone {new_relation}"
                )
                _snowflake_execute_sql(
                    adapter=adapter,
                    sql=six.text_type(create_stmt).strip(),
                    new_conn=False,
                    fetch=False,  # Avoid trying to fetch as pandas
                )
                adapter.drop_relation(new_relation)
            else:
                adapter.rename_relation(new_relation, original_relation)

            adapter.commit_if_has_connection()


def _drop_relation(adapter: SQLAdapter, relation: BaseRelation):
    with adapter.connection_named(_connection_name("drop_relation", relation)):
        with _cache_lock("_drop_relation"):
            adapter.connections.begin()

            _clear_relations_cache(adapter)

            adapter.drop_relation(relation)

            adapter.commit_if_has_connection()


def _alchemy_mock_engine(adapter: SQLAdapter):
    url_string = f"{adapter.type()}://"
    if adapter.type() == "athena":
        SCHEMA_NAME = adapter.config.credentials.schema
        S3_STAGING_DIR = adapter.config.credentials.s3_staging_dir
        AWS_REGION = adapter.config.credentials.region_name

        conn_str = (
            "awsathena+rest://athena.{region_name}.amazonaws.com:443/"
            "{schema_name}?s3_staging_dir={s3_staging_dir}&work_group=primary"
        )

        url_string = conn_str.format(
            region_name=AWS_REGION,
            schema_name=SCHEMA_NAME,
            s3_staging_dir=quote_plus(S3_STAGING_DIR),
        )

    # TODO: add special cases as needed

    def null_dump(sql, *multiparams, **params):
        pass

    return sqlalchemy.create_mock_engine(url_string, executor=null_dump)


def _create_engine_from_connection(adapter: SQLAdapter):
    if adapter.type() == "postgres":
        url_string = "postgresql+psycopg2://"
    else:
        # TODO: add special cases as needed
        LOGGER.warn("No explicit url string for adapter {}", adapter.type())
        url_string = f"{adapter.type()}://"

    connection = adapter.connections.get_thread_connection()
    return sqlalchemy.create_engine(url_string, creator=lambda: connection.handle)


@contextmanager
def _existing_or_new_connection(
    adapter: BaseAdapter,
    name: str,
    new_conn: bool,  # TODO: new_conn solution feels hacky
) -> Iterator[bool]:
    if new_conn:
        with adapter.connection_named(name):
            yield True
    else:
        yield False


# Adapter: salalchemy connection
def _sqlalchemy_engine_fetch_relation(adapter: SQLAdapter, relation: BaseRelation):
    # TODO: use database, just using schema and identifier
    database, schema, identifier = (
        relation.database,
        relation.schema,
        relation.identifier,
    )

    assert identifier

    with _existing_or_new_connection(
        adapter, _connection_name("write_target", relation, _hash=False), True
    ):
        engine = _create_engine_from_connection(adapter)
        # TODO: use database, just using schema and identifier
        return pd.read_sql_table(
            table_name=identifier,
            schema=schema,
            con=engine,
        )


def _sqlalchemy_engine_write_relation(
    adapter: SQLAdapter,
    data: pd.DataFrame,
    relation: BaseRelation,
    *,
    dtype=None,
):
    # TODO: use database, just using schema and identifier
    database, schema, identifier = (
        relation.database,
        relation.schema,
        relation.identifier,
    )

    assert identifier

    with _existing_or_new_connection(
        adapter, _connection_name("write_target", relation, _hash=False), True
    ):
        engine = _create_engine_from_connection(adapter)

        rows_affected = data.to_sql(
            name=identifier,
            con=engine,
            schema=schema,
            if_exists="append",
            index=False,
            dtype=dtype,
        )

        return AdapterResponse("OK", rows_affected=rows_affected)


# Adapter: BigQuery
def _bigquery_execute_sql(
    adapter: BaseAdapter, sql: str, new_conn: bool
) -> Tuple[AdapterResponse, pd.DataFrame]:
    assert adapter.type() == "bigquery"

    import google.cloud.bigquery as bigquery

    with _existing_or_new_connection(
        adapter, _connection_name("bigquery:execute_sql", sql), new_conn
    ):
        connection_manager: BaseConnectionManager = adapter.connections  # type: ignore
        client: bigquery.Client = connection_manager.get_thread_connection().handle  # type: ignore

        job = client.query(sql)
        df = job.to_dataframe()
        if job.destination:
            query_table = client.get_table(job.destination)
            num_rows = query_table.num_rows
        else:
            num_rows = df.size

    # TODO: better AdapterResponse
    return AdapterResponse("OK", rows_affected=num_rows), df


def _bigquery_write_relation(
    adapter: SQLAdapter,
    data: pd.DataFrame,
    relation: BaseRelation,
    *,
    mode: WriteModeEnum,
    fields_schema: Optional[List[dict]] = None,
) -> AdapterResponse:
    import google.cloud.bigquery as bigquery
    from google.cloud.bigquery.job import WriteDisposition
    from dbt.adapters.bigquery import BigQueryAdapter, BigQueryConnectionManager
    from dbt.semver import VersionSpecifier

    assert adapter.type() == "bigquery"

    _adapter: BigQueryAdapter = adapter  # type: ignore

    disposition = (
        WriteDisposition.WRITE_TRUNCATE
        if WriteModeEnum.OVERWRITE == mode
        else WriteDisposition.WRITE_APPEND
    )

    project: str = relation.database  # type: ignore
    dataset: str = relation.schema  # type: ignore
    table: str = relation.identifier  # type: ignore

    with _adapter.connection_named(
        _connection_name("bigquery:write_relation", relation, _hash=False)
    ):
        connection_manager: BigQueryConnectionManager = _adapter.connections
        conn = connection_manager.get_thread_connection()
        client: bigquery.Client = conn.handle  # type: ignore

        table_ref = bigquery.TableReference(
            bigquery.DatasetReference(project, dataset), table
        )

        job_config = bigquery.LoadJobConfig(
            # Specify a (partial) schema. All columns are always written to the
            # table. The schema is used to assist in data type definitions.
            schema=[
                # field_types is a list of API-representation of BigQuery.FieldSchema
                bigquery.SchemaField.from_api_repr(field)
                for field in (fields_schema or [])
            ],
            source_format="PARQUET",
            write_disposition=disposition,
        )

        with connection_manager.exception_handler("START JOB"):
            job = client.load_table_from_dataframe(
                data, table_ref, job_config=job_config
            )

        from dbt.adapters.bigquery.__version__ import version as bigquery_version

        ADAPTER_VCURRENT = VersionSpecifier.from_version_string(bigquery_version)
        # https://github.com/dbt-labs/dbt-bigquery/commit/141b86749df813cf3a3a90a90e7a7dfc401ba9b0
        if ADAPTER_VCURRENT.compare(VersionSpecifier.from_version_string("1.1.0")) >= 0:
            timeout = connection_manager.get_job_execution_timeout_seconds(conn) or 300
        else:
            timeout = connection_manager.get_timeout(conn) or 300

        with connection_manager.exception_handler("LOAD TABLE"):
            _adapter.poll_until_job_completes(job, timeout)

        query_table = client.get_table(job.destination)
        num_rows = query_table.num_rows

    # TODO: better AdapterResponse
    return AdapterResponse("OK", rows_affected=num_rows)


# Adapter: Snowflake
def _snowflake_execute_sql(
    adapter: BaseAdapter,
    sql: str,
    new_conn: bool,
    *,
    fetch: bool = True,
) -> Tuple[AdapterResponse, pd.DataFrame]:
    assert adapter.type() == "snowflake"

    import snowflake.connector as snowflake
    from dbt.adapters.snowflake import SnowflakeConnectionManager

    with _existing_or_new_connection(
        adapter, _connection_name("snowflake:execute_sql", sql), new_conn
    ):
        connection_manager: SnowflakeConnectionManager = adapter.connections  # type: ignore
        conn: snowflake.SnowflakeConnection = connection_manager.get_thread_connection().handle  # type: ignore

        with connection_manager.exception_handler("EXECUTE SQL"):
            cur = conn.cursor()

            cur.execute(sql)

            # Use snowflake-dbt function directly
            res = connection_manager.get_response(cur)

            df = pd.DataFrame({})
            if fetch:
                df: pd.DataFrame = cur.fetch_pandas_all()

                # HACK: manually parse ARRAY and VARIANT since they are returned as strings right now
                # Related issue: https://github.com/snowflakedb/snowflake-connector-python/issues/544
                for desc in cur.description:
                    # 5=VARIANT, 10=ARRAY -- https://docs.snowflake.com/en/user-guide/python-connector-api.html#type-codes
                    if desc.type_code in [5, 10]:
                        import json

                        df[desc.name] = df[desc.name].map(lambda v: json.loads(v))

    return res, df


def _snowflake_write_relation(
    adapter: SQLAdapter,
    data: pd.DataFrame,
    relation: BaseRelation,
) -> AdapterResponse:
    from dbt.adapters.snowflake import SnowflakeAdapter, SnowflakeConnectionManager
    import snowflake.connector as snowflake
    import snowflake.connector.pandas_tools as snowflake_pandas

    assert adapter.type() == "snowflake"

    _adapter: SnowflakeAdapter = adapter  # type: ignore

    database: str = relation.database  # type: ignore
    schema: str = relation.schema  # type: ignore
    table: str = relation.identifier  # type: ignore

    with _adapter.connection_named(
        _connection_name("snowflake:write_relation", relation, _hash=False)
    ):
        connection_manager: SnowflakeConnectionManager = _adapter.connections  # type: ignore
        conn: snowflake.SnowflakeConnection = connection_manager.get_thread_connection().handle  # type: ignore

        with connection_manager.exception_handler("LOAD TABLE"):
            success, chunks, num_rows, output = snowflake_pandas.write_pandas(
                conn,
                data,
                table_name=table,
                database=database,
                schema=schema,
                auto_create_table=True,
                quote_identifiers=False,
            )
            if not success:
                # In case the failure does not raise by itself
                # I have not been able to reproduce such a case
                from dbt.exceptions import DatabaseException

                raise DatabaseException(output)

    # TODO: better AdapterResponse
    return AdapterResponse(str(output), rows_affected=num_rows)



from contextlib import contextmanager
from typing import Any

from dbt.config.runtime import RuntimeConfig

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache


FAL_SCRIPTS_PATH_VAR_NAME = 'fal-scripts-path'


def cache_static(func):
    """Cache the result of a function."""
    return lru_cache(maxsize=None)(func)


def retrieve_symbol(source_code: str, symbol_name: str) -> Any:
    """Retrieve the function with the given name from the source code."""
    namespace = {}
    exec(source_code, namespace)
    return namespace[symbol_name]


def get_fal_scripts_path(config: RuntimeConfig):
    import pathlib
    project_path = pathlib.Path(config.project_root)

    # Default value
    fal_scripts_path = ''

    if hasattr(config, 'vars'):
        fal_scripts_path: str = config.vars.to_dict().get(FAL_SCRIPTS_PATH_VAR_NAME, fal_scripts_path)  # type: ignore

    if hasattr(config, 'cli_vars'):
        fal_scripts_path = config.cli_vars.get(FAL_SCRIPTS_PATH_VAR_NAME, fal_scripts_path)

    return project_path / fal_scripts_path


@contextmanager
def extra_path(path: str):
    import sys
    sys.path.append(path)
    try:
        yield
    finally:
        sys.path.remove(path)



from dbt.contracts.results import NodeStatus

from fal.dbt.integration.project import (
    FalDbt,
    DbtModel,
    DbtSource,
    DbtTest,
    DbtGenericTest,
    DbtSingularTest,
)
from fal.dbt.fal_script import Context, CurrentModel



from __future__ import annotations

import zipfile
import io
import functools
import pandas as pd
from functools import partial
from tempfile import NamedTemporaryFile
from typing import Any, Callable, Dict, NewType, Optional

from dbt.config.runtime import RuntimeConfig
from dbt.contracts.connection import AdapterResponse
from dbt.flags import get_flags, Namespace

from fal_serverless import FalServerlessHost, isolated
from dbt.adapters.fal_experimental.connections import TeleportTypeEnum
from dbt.adapters.fal_experimental.utils.environments import (
    EnvironmentDefinition,
    get_default_pip_dependencies,
)
from dbt.adapters.fal_experimental.utils import (
    extra_path,
    get_fal_scripts_path,
    retrieve_symbol,
)
from dbt.fal.adapters.teleport.info import TeleportInfo


DataLocation = NewType("DataLocation", Dict[str, str])


def _prepare_for_teleport(
    function: Callable, teleport: TeleportInfo, locations: DataLocation
) -> Callable:
    @functools.wraps(function)
    def wrapped(relation: str, *args, **kwargs) -> Any:
        relation = relation.lower()
        return function(teleport, locations, relation, *args, **kwargs)

    return wrapped


def _teleport_df_from_external_storage(
    teleport_info: TeleportInfo, locations: DataLocation, relation: str
) -> pd.DataFrame:
    if relation not in locations:
        raise RuntimeError(f"Could not find url for '{relation}' in {locations}")

    if teleport_info.format == "parquet":
        relation_path = locations[relation]
        url = teleport_info.build_url(relation_path)
        storage_options = _build_teleport_storage_options(teleport_info)
        return pd.read_parquet(url, storage_options=storage_options)
    else:
        # TODO: support more
        raise RuntimeError(f"Format {teleport_info.format} not supported")


def _teleport_df_to_external_storage(
    teleport_info: TeleportInfo,
    locations: DataLocation,
    relation: str,
    data: pd.DataFrame,
):
    if teleport_info.format == "parquet":
        relation_path = teleport_info.build_relation_path(relation)
        url = teleport_info.build_url(relation_path)
        storage_options = _build_teleport_storage_options(teleport_info)

        data.to_parquet(url, storage_options=storage_options)
        locations[relation] = relation_path
        return relation_path
    else:
        raise RuntimeError(f"Format {teleport_info.format} not supported")


def _build_teleport_storage_options(teleport_info: TeleportInfo) -> Dict[str, str]:
    storage_options = {}
    if teleport_info.credentials.type == TeleportTypeEnum.REMOTE_S3:
        storage_options = {
            "key": teleport_info.credentials.s3_access_key_id,
            "secret": teleport_info.credentials.s3_access_key,
        }
    elif teleport_info.credentials.type == TeleportTypeEnum.LOCAL:
        pass
    else:
        raise RuntimeError(
            f"Teleport storage type {teleport_info.credentials.type} not supported"
        )
    return storage_options


def run_with_teleport(
    code: str,
    teleport_info: TeleportInfo,
    locations: DataLocation,
    config: RuntimeConfig,
    local_packages: Optional[bytes] = None,
) -> str:
    # main symbol is defined during dbt-fal's compilation
    # and acts as an entrypoint for us to run the model.
    fal_scripts_path = str(get_fal_scripts_path(config))
    if local_packages is not None:
        if fal_scripts_path.exists():
            import shutil

            shutil.rmtree(fal_scripts_path)
        fal_scripts_path.parent.mkdir(parents=True, exist_ok=True)
        zip_file = zipfile.ZipFile(io.BytesIO(local_packages))
        zip_file.extractall(fal_scripts_path)

    with extra_path(fal_scripts_path):
        main = retrieve_symbol(code, "main")
        return main(
            read_df=_prepare_for_teleport(
                _teleport_df_from_external_storage, teleport_info, locations
            ),
            write_df=_prepare_for_teleport(
                _teleport_df_to_external_storage, teleport_info, locations
            ),
        )


def run_in_environment_with_teleport(
    environment: EnvironmentDefinition,
    code: str,
    teleport_info: TeleportInfo,
    locations: DataLocation,
    config: RuntimeConfig,
    adapter_type: str,
) -> AdapterResponse:
    """Run the 'main' function inside the given code on the
    specified environment.

    The environment_name must be defined inside fal_project.yml file
    in your project's root directory."""
    compressed_local_packages = None
    is_remote = type(environment.host) is FalServerlessHost

    deps = get_default_pip_dependencies(
        is_remote=is_remote,
        adapter_type=adapter_type,
        is_teleport=True,
    )

    fal_scripts_path = get_fal_scripts_path(config)

    if is_remote and fal_scripts_path.exists():
        with NamedTemporaryFile() as temp_file:
            with zipfile.ZipFile(temp_file.name, "w", zipfile.ZIP_DEFLATED) as zip_file:
                for entry in fal_scripts_path.rglob("*"):
                    zip_file.write(entry, entry.relative_to(fal_scripts_path))

            compressed_local_packages = temp_file.read()

    execute_model = partial(
        run_with_teleport,
        code=code,
        teleport_info=teleport_info,
        locations=locations,
        config=config,
        local_packages=compressed_local_packages,
    )

    if environment.kind == "virtualenv":
        requirements = environment.config.get("requirements", [])
        requirements += deps
        isolated_function = isolated(
            kind="virtualenv", host=environment.host, requirements=requirements
        )(execute_model)
    elif environment.kind == "conda":
        dependencies = environment.config.pop("packages", [])
        dependencies.append({"pip": deps})
        env_dict = {
            "name": "dbt_fal_env",
            "channels": ["conda-forge", "defaults"],
            "dependencies": dependencies,
        }
        isolated_function = isolated(
            kind="conda", host=environment.host, env_dict=env_dict
        )(execute_model)
    else:
        # We should not reach this point, because environment types are validated when the
        # environment objects are created (in utils/environments.py).
        raise Exception(f"Environment type not supported: {environment.kind}")

    # Machine type is only applicable in FalServerlessHost
    if is_remote:
        isolated_function = isolated_function.on(machine_type=environment.machine_type)

    result = isolated_function()
    return result



# TODO: remove once `fal` is no longer a supported package
# The whole file
DBT_FAL_IMPORT_NOTICE = \
"""The dbt tool `fal` and `dbt-fal` adapter have been merged into a single tool.
Please import from the `fal.dbt` module instead.
Running `pip install dbt-fal` will install the new tool and the adapter alongside.
Then import from the `fal.dbt` module like

    from fal.dbt import {name}
"""

global DBT_FAL_IMPORT_NOTICE_SHOWN
DBT_FAL_IMPORT_NOTICE_SHOWN = {}

def _warn(name: str):
    # Print once per deprecated import
    global DBT_FAL_IMPORT_NOTICE_SHOWN
    if not DBT_FAL_IMPORT_NOTICE_SHOWN.get(name, False):
        print(DBT_FAL_IMPORT_NOTICE.format(name=name))
        DBT_FAL_IMPORT_NOTICE_SHOWN[name] = True

# Avoid printing on non-direct imports
def __getattr__(name: str):
    if name == "NodeStatus":
        _warn(name)
        from fal.dbt import NodeStatus
        return NodeStatus
    elif name == "FalDbt":
        _warn(name)
        from fal.dbt import FalDbt
        return FalDbt
    elif name == "DbtModel":
        _warn(name)
        from fal.dbt import DbtModel
        return DbtModel
    elif name == "DbtSource":
        _warn(name)
        from fal.dbt import DbtSource
        return DbtSource
    elif name == "DbtTest":
        _warn(name)
        from fal.dbt import DbtTest
        return DbtTest
    elif name == "DbtGenericTest":
        _warn(name)
        from fal.dbt import DbtGenericTest
        return DbtGenericTest
    elif name == "DbtSingularTest":
        _warn(name)
        from fal.dbt import DbtSingularTest
        return DbtSingularTest
    elif name == "Context":
        _warn(name)
        from fal.dbt import Context
        return Context
    elif name == "CurrentModel":
        _warn(name)
        from fal.dbt import CurrentModel
        return CurrentModel

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from contextlib import contextmanager
import logging
import datetime
import sys

import dbt.ui as ui
from dbt.logger import log_manager as dbt_log_manager

TRACE = logging.DEBUG - 5
logging.addLevelName(TRACE, "TRACE")

class FalLogger:
    def __init__(self):
        self._stdout_handler = logging.StreamHandler(sys.stdout)
        self._stdout_handler.setLevel(logging.INFO)

        self._logger = logging.Logger("fal", logging.INFO)
        self._logger.addHandler(self._stdout_handler)

    def __getstate__(self):
        # Don't pickle the logger
        d = self.__dict__.copy()
        d['_logger'] = d['_logger'].name
        return d

    def __setstate__(self, d):
        # Set logger when unpickling
        d['_logger'] = logging.Logger(
            d['_logger'] if isinstance(d['_logger'], str) else "fal", logging.INFO)
        self.__dict__.update(d)

    def set_level(self, level: int):
        self._logger.setLevel(level)

    @property
    def level(self):
        return self._logger.level

    def log(self, level: int, msg: str, *args, **kwargs):
        now = datetime.datetime.now(datetime.timezone.utc)
        if self.level <= logging.DEBUG:
            prefix = now.strftime(r"%H:%M:%S.%f")

            prefix += f" [{logging.getLevelName(level).lower()[0:5].ljust(5)}]"

            # Spaces to match the spacing of dbt's debug logs, like the following
            #
            #     21:32:31.816189 [debug] [MainThread]: Flushing usage events
            #     21:32:32.530385 [error] [fal       ]: Error in script (...):
            #     21:32:34.554038 [debug] [Thread-1  ]: Opening a new connection, currently in state closed
            prefix += " [fal       ]:"
        else:
            prefix = now.strftime("%H:%M:%S")
            prefix += " [fal]:"

        self._logger.log(level, f"{prefix} {_prepare_msg(msg, *args, **kwargs)}")

    def trace(self, msg: str, *args, **kwargs):
        self.log(TRACE, msg, *args, **kwargs)

    def debug(self, msg: str, *args, **kwargs):
        self.log(logging.DEBUG, msg, *args, **kwargs)

    def info(self, msg: str, *args, **kwargs):
        self.log(logging.INFO, msg, *args, **kwargs)

    def warning(self, msg: str, *args, **kwargs):
        self.log(logging.WARNING, ui.warning_tag(msg), *args, **kwargs)

    def warn(self, msg: str, *args, **kwargs):
        # Alias to warning
        return self.warning(msg, *args, **kwargs)

    def error(self, msg: str, *args, **kwargs):
        self.log(logging.ERROR, msg, *args, **kwargs)

class LogManager():
    def __init__(self, dbt_log_manager):
        self._dbt_log_manager = dbt_log_manager

    @contextmanager
    def applicationbound(self):
        # TODO: probably where we can add threadding information if we want it for debug logs
        with self._dbt_log_manager.applicationbound():
            yield

    def set_debug(self):
        self._dbt_log_manager.set_debug()
        LOGGER.set_level(logging.DEBUG)

    def set_trace(self):
        self._dbt_log_manager.set_debug()
        LOGGER.set_level(TRACE)

def _prepare_msg(msg: str, *args, **kwargs):
    if args or kwargs:
        return msg.format(*args, **kwargs)
    else:
        return msg

LOGGER = FalLogger()
log_manager = LogManager(dbt_log_manager)



# A Python native two-way bridge implementation between the controller
# (fal-cli) process and the child process(es) (isolated scripts).
#
#    controller       both       children
# /--------------\ /--------\ /------------\
# object -> dill -> socket -> dill -> object
#
#
# The controller starts the listener (server) and then passes the raw
# socket address to the children. Each child then connects to the server
# and from that point onward, all exchanges (via send/recv) are done via
# regular Python objects (using dill).
#


from __future__ import annotations

import base64
from contextlib import closing
from functools import partial
from multiprocessing.connection import Client, ConnectionWrapper, Listener
from typing import Union

import dill

dill_wrapper = partial(ConnectionWrapper, dumps=dill.dumps, loads=dill.loads)


class _DillListener(Listener):
    def accept(self) -> ConnectionWrapper:
        return closing(dill_wrapper(super().accept()))


def controller_connection() -> _DillListener:
    # The controller here assumes that there will be at most one
    # client. This restriction might change in the future as an
    # optimization.
    return _DillListener()


def child_connection(address: str) -> ConnectionWrapper:
    return closing(dill_wrapper(Client(address)))


def encode_service_address(address: Union[bytes, str]) -> str:
    if isinstance(address, bytes):
        address = address.decode()

    return base64.b64encode(address.encode()).decode("utf-8")


def decode_service_address(address: str) -> str:
    return base64.b64decode(address).decode("utf-8")



import dbt.version
from dbt.semver import VersionSpecifier

def version_compare(version_string: str):
    return DBT_VCURRENT.compare(VersionSpecifier.from_version_string(version_string))

def is_version_plus(version_string: str):
    return version_compare(version_string) >= 0

DBT_VCURRENT = dbt.version.installed



from typing import Dict, Any, Tuple, Optional

from dbt.flags import get_flags, Namespace
from dbt.config.project import load_raw_project
from dbt.config.profile import read_profile, Profile
from dbt.config.renderer import ProfileRenderer


# NOTE: inspired in https://github.com/dbt-labs/dbt-core/blob/828d723512fced809c63e369a82c7eb570a74986/core/dbt/config/runtime.py#L58-L74
def find_profile_name(
    profile_override: Optional[str],
    project_root: str,
    profile_renderer: ProfileRenderer,
):
    if profile_override is not None:
        profile_name = profile_override
    else:
        raw_project = load_raw_project(project_root)
        raw_profile_name = raw_project.get("profile")
        profile_name = profile_renderer.render_value(raw_profile_name)

    return profile_name


# NOTE: inspired in https://github.com/dbt-labs/dbt-core/blob/828d723512fced809c63e369a82c7eb570a74986/core/dbt/config/profile.py#L279-L311
def find_target_name(
    target_override: Optional[str], raw_profile: dict, profile_renderer: ProfileRenderer
):
    if target_override is not None:
        target_name = target_override
    elif "target" in raw_profile:
        # render the target if it was parsed from yaml
        target_name = profile_renderer.render_value(raw_profile["target"])
    else:
        target_name = "default"

    return target_name


def load_profiles_info_1_5() -> Tuple[Profile, Dict[str, Any]]:
    flags: Namespace = get_flags()

    profile_renderer = ProfileRenderer(getattr(flags, "VARS", {}))

    profile_name = find_profile_name(flags.PROFILE, flags.PROJECT_DIR, profile_renderer)

    raw_profiles = read_profile(flags.PROFILES_DIR)
    raw_profile = raw_profiles[profile_name]

    target_name = find_target_name(flags.TARGET, raw_profile, profile_renderer)

    fal_dict = Profile._get_profile_data(
        profile=raw_profile,
        profile_name=profile_name,
        target_name=target_name,
    )
    db_profile_target_name = fal_dict.get("db_profile")
    assert (
        db_profile_target_name
    ), "fal credentials must have a `db_profile` property set"

    try:
        db_profile = Profile.from_raw_profile_info(
            raw_profile=raw_profile,
            profile_name=profile_name,
            renderer=profile_renderer,
            # TODO: should we load the user_config?
            user_config={},
            target_override=db_profile_target_name,
        )
    except RecursionError as error:
        raise AttributeError(
            "Did you wrap a type 'fal' profile with another type 'fal' profile?"
        ) from error

    override_properties = {
        "threads": getattr(flags, "THREADS", None) or fal_dict.get("threads") or db_profile.threads,
    }

    return db_profile, override_properties



from dbt.adapters.base import AdapterPlugin

from dbt.adapters.fal.connections import FalEncCredentials
from dbt.include import fal

# Avoid loading the plugin code for any import
def __getattr__(name):
    if name == "FalEncAdapter":
        from dbt.adapters.fal.impl import FalEncAdapter

        return FalEncAdapter
    if name == "Plugin":
        from dbt.adapters.fal.impl import FalEncAdapter

        return AdapterPlugin(
            adapter=FalEncAdapter,
            credentials=FalEncCredentials,
            include_path=fal.PACKAGE_PATH
        )

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from typing import List
import sys

from click.exceptions import ClickException

from fal.dbt.cli.flow_runner import fal_flow_run
from .args import parse_args
from .fal_runner import fal_run
from fal.dbt.telemetry import telemetry

from fal.dbt.integration.logger import log_manager


# TODO: remove once `fal` is no longer a supported package
DBT_FAL_COMMAND_NOTICE = \
"""The dbt tool `fal` and `dbt-fal` adapter have been merged into a single tool.
Please use the new `dbt-fal` command line tool instead.
Running `pip install dbt-fal` will install the new tool and the adapter alongside.
Then run your command like

    dbt-fal <command>
"""

# TODO: remove once `fal` is no longer a supported package
def fal_cli(argv: List[str] = sys.argv):
    print(DBT_FAL_COMMAND_NOTICE)
    cli(argv)

def cli(argv: List[str] = sys.argv):
    # Wrapper to be able to shutdown telemetry afterwards
    try:
        _cli(argv)
    finally:
        telemetry.shutdown()


@telemetry.log_call("cli")
def _cli(argv: List[str]):
    parsed = parse_args(argv[1:])

    # TODO: do we still need this?
    with log_manager.applicationbound():
        if parsed.debug:
            log_manager.set_debug()

        if parsed.command == "flow":
            if parsed.flow_command == "run":
                exit_code = fal_flow_run(parsed)
                if exit_code:
                    raise SystemExit(exit_code)

        elif parsed.command == "run":
            fal_run(parsed)

        else:
            raise ClickException(f"Unknown command {parsed.command}")



import ast
from functools import partial
import re
from typing import Callable, Iterable, List, TypeVar
from pathlib import Path
from fal.dbt.fal_script import python_from_file

from fal.dbt.integration.parse import get_fal_models_dirs, load_dbt_project_contract
from fal.dbt.cli.model_generator.module_check import (
    generate_dbt_dependencies,
    write_to_model_check,
)

from fal.dbt.integration.logger import LOGGER

from fal.dbt.telemetry import telemetry

SQL_MODEL_TEMPLATE = """
{{ config(materialized='ephemeral') }}
/*
FAL_GENERATED __checksum__

Script dependencies:

__deps__

*/

SELECT * FROM {{ this }}
"""

GENERATED_DIR = Path("fal")

CHECKSUM_REGEX = re.compile(r"FAL_GENERATED ([_\d\w]+)")


def generate_python_dbt_models(project_dir: str, args_vars: str):
    fal_models_paths = _get_fal_models_paths(project_dir, args_vars)
    dbt_models_paths = list(
        map(Path, load_dbt_project_contract(project_dir).model_paths or [])
    )

    base_dbt_models_dir = Path(project_dir, dbt_models_paths[0])

    old_generated_sqls = list(
        _process_models_paths(dbt_models_paths, _find_fal_generated_models)
    )

    fal_python_models_and_sqls = list(
        _process_models_paths(
            fal_models_paths,
            partial(_find_fal_python_models_and_sql_target, base_dbt_models_dir),
        )
    )

    python_paths: List[Path] = []
    fal_target_sqls: List[Path] = []
    if fal_python_models_and_sqls:
        python_paths, fal_target_sqls, _ = map(list, zip(*fal_python_models_and_sqls))

    _delete_old_generated_sqls(old_generated_sqls, fal_target_sqls)

    for py_path, sql_path, old_checksum in fal_python_models_and_sqls:
        new_checksum = _generate_sql_for_fal_python_model(py_path, sql_path)
        if not old_checksum or new_checksum != old_checksum:
            LOGGER.warn(
                f"File '{sql_path.relative_to(project_dir)}' was generated from '{py_path.relative_to(project_dir)}'.\n"
                "Please do not modify it directly. We recommend committing it to your repository."
            )

    if python_paths:
        telemetry.log_api(
            action="python_models_generated",
            additional_props={"models": len(python_paths)},
        )

    return {path.stem: path for path in python_paths}


def _get_fal_models_paths(project_dir: str, args_vars: str):
    models_paths = get_fal_models_dirs(project_dir, args_vars)
    project_path = Path(project_dir)
    return list(map(project_path.joinpath, models_paths))


def _generate_sql_for_fal_python_model(py_path: Path, sql_path: Path):
    source_code = python_from_file(py_path)
    module = ast.parse(source_code, str(py_path), "exec")

    # Fails if it does not have write_to_model
    write_to_model_check(module)

    dbt_deps = generate_dbt_dependencies(module)

    sql_contents = SQL_MODEL_TEMPLATE.replace("__deps__", dbt_deps)
    checksum, _ = _checksum(sql_contents)
    sql_contents = sql_contents.replace("__checksum__", checksum)

    sql_path.parent.mkdir(parents=True, exist_ok=True)
    with open(sql_path, "w") as file:
        file.write(sql_contents)

    return checksum


# TODO: unit tests
def _check_path_safe_to_write(sql_path: Path, py_path: Path):
    if sql_path.exists():
        with open(sql_path, "r") as file:
            contents = file.read()
            checksum, found = _checksum(contents)
            if not found or checksum != found:
                LOGGER.debug(
                    f"Existing file calculated checksum: {checksum}\nFound checksum: {found}"
                )
                raise RuntimeError(
                    f"File '{sql_path}' not generated by fal would be "
                    f"overwritten by generated model of '{py_path}'. Please rename or remove."
                )
            return checksum


T = TypeVar("T")


def _process_models_paths(
    models_paths: Iterable[Path], func: Callable[[Path], Iterable[T]]
) -> Iterable[T]:
    for models_path in models_paths:
        yield from func(models_path)


def _find_fal_python_models_and_sql_target(base_sql_path: Path, models_path: Path):
    for py_path in _find_python_files(models_path):
        sql_path = _sql_path_from_python_path(
            base_sql_path, py_path.relative_to(models_path)
        )
        old_checksum = _check_path_safe_to_write(sql_path, py_path)
        yield py_path, sql_path, old_checksum


def _sql_path_from_python_path(base_sql_path: Path, relative_py_path: Path):
    return base_sql_path / GENERATED_DIR / relative_py_path.with_suffix(".sql")


def _find_fal_generated_models(models_path: Path):
    fal_path = models_path / GENERATED_DIR
    return (file for file in fal_path.glob("**/*.sql") if _is_fal_generated(file))


def _delete_old_generated_sqls(old_models: Iterable[Path], new_models: Iterable[Path]):
    for sql_path in old_models:
        if sql_path not in new_models:
            sql_path.unlink()


def _is_fal_generated(file_path):
    with open(file_path) as file:
        return CHECKSUM_REGEX.search(file.read())


def _checksum(contents: str):
    import hashlib

    found = CHECKSUM_REGEX.search(contents)
    to_check = CHECKSUM_REGEX.sub("FAL_GENERATED", contents.strip())
    return (
        hashlib.md5(to_check.encode("utf-8")).hexdigest(),
        found.group(1) if found else None,
    )


def _find_python_files(models_path: Path) -> List[Path]:
    files = []
    files.extend(models_path.rglob("*.py"))
    files.extend(models_path.rglob("*.ipynb"))

    return [p for p in files if p.is_file()]



import csv
from io import StringIO

import pandas as pd
import sqlalchemy

from dbt.adapters.base import BaseRelation
from dbt.adapters.base.connections import AdapterResponse
from dbt.adapters.fal_experimental.adapter_support import drop_relation_if_it_exists, new_connection
from dbt.adapters.postgres import PostgresAdapter


def read_relation_as_df(
    adapter: PostgresAdapter, relation: BaseRelation
) -> pd.DataFrame:
    assert adapter.type() == "postgres"

    with new_connection(adapter, "fal-postgres:read_relation_as_df") as connection:
        # If the given adapter supports the DBAPI (PEP 249), we can
        # use its connection directly for the engine.
        alchemy_engine = sqlalchemy.create_engine(
            "postgresql+psycopg2://",
            creator=lambda *args, **kwargs: connection.handle,
        )

        return pd.read_sql_table(
            con=alchemy_engine,
            table_name=relation.identifier,
            schema=relation.schema,
        )


def write_df_to_relation(
    adapter: PostgresAdapter,
    data: pd.DataFrame,
    relation: BaseRelation,
    *,
    if_exists: str = "replace",
) -> AdapterResponse:
    assert adapter.type() == "postgres"

    with new_connection(adapter, "fal-postgres:write_df_to_relation") as connection:
        # TODO: this should probably live in the materialization macro.
        temp_relation = relation.replace_path(
            identifier=f"__dbt_fal_temp_{relation.identifier}"
        )
        drop_relation_if_it_exists(adapter, temp_relation)

        alchemy_engine = sqlalchemy.create_engine(
            "postgresql+psycopg2://",
            creator=lambda *args, **kwargs: connection.handle,
        )

        # TODO: probably worth handling errors here an returning
        # a proper adapter response.
        rows_affected = data.to_sql(
            con=alchemy_engine,
            name=temp_relation.identifier,
            schema=temp_relation.schema,
            if_exists=if_exists,
            index=False,
            method=_psql_insert_copy,
        )
        adapter.cache.add(temp_relation)
        drop_relation_if_it_exists(adapter, relation)
        adapter.rename_relation(temp_relation, relation)
        adapter.commit_if_has_connection()

        return AdapterResponse("OK", rows_affected=rows_affected)


def _psql_insert_copy(table, conn, keys, data_iter):
    """Alternative to_sql method for PostgreSQL.

    Adapted from https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html#io-sql-method
    """
    dbapi_conn = conn.connection
    with dbapi_conn.cursor() as cur:
        s_buf = StringIO()
        writer = csv.writer(s_buf)
        writer.writerows(data_iter)
        s_buf.seek(0)

        columns = ", ".join((f'"{k}"' for k in keys))
        table_name = f"{table.schema}.{table.name}" if table.schema else table.name

        sql = f"COPY {table_name} ({columns}) FROM STDIN WITH CSV"
        cur.copy_expert(sql=sql, file=s_buf)



from __future__ import annotations
from dataclasses import dataclass
from typing import List, Tuple, Dict, cast

from fal.dbt.fal_script import FalScript, TimingType
from fal.dbt.integration.project import DbtModel, FalDbt
from pathlib import Path
import networkx as nx
import os as os
from functools import reduce
from enum import Enum


class NodeKind(str, Enum):
    DBT_MODEL = "dbt model"
    FAL_MODEL = "fal model"
    FAL_SCRIPT = "fal script"


@dataclass
class FalFlowNode:
    "Represents a Node that can be invoked by fal flow command"
    unique_id: str


@dataclass
class ScriptNode(FalFlowNode):
    "Represents a python script node"
    script: FalScript


@dataclass
class DbtModelNode(FalFlowNode):
    "Represents a dbt node"
    model: DbtModel


def _add_after_scripts(
    model: DbtModel,
    upstream_fal_node_unique_id: str,
    faldbt: FalDbt,
    graph: nx.DiGraph,
    nodeLookup: Dict[str, FalFlowNode],
) -> Tuple[nx.DiGraph, Dict[str, FalFlowNode]]:
    "Add dbt node to after scripts edges to the graph"
    after_scripts = model.get_scripts(before=False)
    after_fal_scripts = map(
        lambda script_path: FalScript(faldbt, model, script_path, timing_type=TimingType.POST), after_scripts
    )
    after_fal_script_nodes = list(
        map(
            lambda fal_script: ScriptNode(
                _script_id_from_path(fal_script.path, model.name, "AFTER"), fal_script
            ),
            after_fal_scripts,
        )
    )
    for fal_script_node in after_fal_script_nodes:
        graph.add_node(fal_script_node.unique_id, kind=NodeKind.FAL_SCRIPT)
        nodeLookup[fal_script_node.unique_id] = fal_script_node
        # model_fal_node depends on fal_script_node
        graph.add_edge(upstream_fal_node_unique_id, fal_script_node.unique_id)

    return graph, nodeLookup


def _add_before_scripts(
    model: DbtModel,
    downstream_fal_node_unique_id: str,
    faldbt: FalDbt,
    graph: nx.DiGraph,
    nodeLookup: Dict[str, FalFlowNode],
) -> Tuple[nx.DiGraph, Dict[str, FalFlowNode]]:
    "Add before scripts to dbt node edges to the graph"
    before_scripts = model.get_scripts(before=True)
    before_fal_scripts = map(
        lambda script_path: FalScript(faldbt, model, script_path, timing_type=TimingType.PRE), before_scripts
    )
    before_fal_script_node = map(
        lambda fal_script: ScriptNode(
            _script_id_from_path(fal_script.path, model.name, "BEFORE"), fal_script
        ),
        before_fal_scripts,
    )

    for fal_script_node in before_fal_script_node:
        graph.add_node(fal_script_node.unique_id, kind=NodeKind.FAL_SCRIPT)
        nodeLookup[fal_script_node.unique_id] = fal_script_node
        # fal_script_node depends on model_fal_node
        graph.add_edge(fal_script_node.unique_id, downstream_fal_node_unique_id)

    return graph, nodeLookup


def _script_id_from_path(scriptPath: Path, modelName: str, order: str):
    script_file_name = os.path.basename(scriptPath)
    return f"script.{modelName}.{order}.{script_file_name}"


@dataclass
class NodeGraph:
    "Wrapper around networkx graph"

    @classmethod
    def from_fal_dbt(cls, fal_dbt: FalDbt):
        graph = nx.DiGraph()
        node_lookup: Dict[str, FalFlowNode] = {}
        for model in cast(List[DbtModel], fal_dbt.list_models()):
            model_fal_node = DbtModelNode(model.unique_id, model)
            node_lookup[model_fal_node.unique_id] = model_fal_node
            graph.add_node(
                model_fal_node.unique_id,
                kind=NodeKind.FAL_MODEL if model.python_model else NodeKind.DBT_MODEL,
                pre_hook=model.get_hooks(TimingType.PRE),
                post_hook=model.get_hooks(TimingType.POST),
                environment=model.environment_name,
            )

            # Add dbt model dependencies
            for dbt_dependency_unique_id in model_fal_node.model.get_depends_on_nodes():
                if dbt_dependency_unique_id not in node_lookup:
                    graph.add_node(dbt_dependency_unique_id, kind=NodeKind.DBT_MODEL)
                graph.add_edge(dbt_dependency_unique_id, model_fal_node.unique_id)

            _add_after_scripts(
                model,
                model_fal_node.unique_id,
                fal_dbt,
                graph,
                node_lookup,
            )

            _add_before_scripts(
                model,
                model_fal_node.unique_id,
                fal_dbt,
                graph,
                node_lookup,
            )

        return cls(graph, node_lookup)

    def __init__(self, graph: nx.DiGraph, node_lookup: Dict[str, FalFlowNode]):
        self.graph = graph
        self.node_lookup = node_lookup

    def get_successors(self, id: str, levels: int) -> List[str]:
        assert levels >= 0
        if levels == 0:
            return []
        else:
            current: List[str] = list(self.graph.successors(id))
            return reduce(
                lambda acc, id: acc + self.get_successors(id, levels - 1),
                current,
                current,
            )

    def get_descendants(self, id: str) -> List[str]:
        return list(nx.descendants(self.graph, id))

    def get_predecessors(self, id: str, levels: int) -> List[str]:
        assert levels >= 0
        if levels == 0:
            return []
        else:
            current: List[str] = list(self.graph.predecessors(id))
            return reduce(
                lambda acc, id: acc + self.get_predecessors(id, levels - 1),
                current,
                current,
            )

    def get_ancestors(self, id: str) -> List[str]:
        return list(nx.ancestors(self.graph, id))

    def get_node(self, id: str) -> FalFlowNode | None:
        return self.node_lookup.get(id)

    def _is_script_node(self, node_name: str) -> bool:
        return _is_script(node_name)

    def _is_critical_node(self, node):
        successors = list(self.graph.successors(node))
        if node in successors:
            successors.remove(node)

        def is_python_model(id: str):
            inode = self.get_node(id)
            if isinstance(inode, DbtModelNode):
                return inode.model.python_model
            return False

        def has_post_hooks(node_id: str):
            inode = self.get_node(node_id)
            if isinstance(inode, DbtModelNode):
                return bool(inode.model.get_hooks(TimingType.POST))
            return False

        is_model_pred = lambda node_name: node_name.split(".")[0] == "model"
        # fmt:off
        return (
            (any(_is_script(i) for i in successors) and
            any(is_model_pred(i) for i in successors)) or
            is_python_model(node) or
            has_post_hooks(node)
        )
        # fmt:on

    def sort_nodes(self):
        return nx.topological_sort(self.graph)

    def _group_nodes(self) -> List[List[str]]:
        nodes = list(self.sort_nodes())
        buckets = []
        local_bucket = []
        seen_nodes = []
        for node in nodes:
            if node not in seen_nodes:
                local_bucket.append(node)
                seen_nodes.append(node)
                if self._is_critical_node(node):
                    script_successors = list(
                        filter(
                            _is_script,
                            self.graph.successors(node),
                        )
                    )
                    seen_nodes.extend(script_successors)
                    local_bucket.extend(script_successors)
                    buckets.append(local_bucket)
                    local_bucket = []
        buckets.append(local_bucket)
        return buckets

    def generate_sub_graphs(self) -> List[NodeGraph]:
        "Generates subgraphs that are seperated by `critical nodes`"
        sub_graphs = []
        for bucket in self._group_nodes():
            sub_graph = self.graph.subgraph(bucket)
            sub_graph_nodes = list(sub_graph.nodes())
            local_lookup = dict(
                filter(
                    lambda node: node[0] in sub_graph_nodes, self.node_lookup.items()
                )
            )
            node_graph = NodeGraph(sub_graph, local_lookup)
            sub_graphs.append(node_graph)
        return sub_graphs


def _is_script(name: str) -> bool:
    return name.endswith(".py") or name.endswith(".ipynb")



from __future__ import annotations

import sys
import hashlib
import os
import shutil
import subprocess
import sysconfig
from dataclasses import dataclass
from pathlib import Path
from typing import List, Dict, Any

from fal.dbt.packages.environments.base import (
    BASE_CACHE_DIR,
    BaseEnvironment,
    DualPythonIPC,
    log_env,
    rmdir_on_fail,
)
from fal.dbt.packages.environments.virtual_env import get_primary_virtual_env
from fal.dbt.utils import cache_static

_BASE_CONDA_DIR = BASE_CACHE_DIR / "conda"
_BASE_CONDA_DIR.mkdir(exist_ok=True)

# Specify the path where the conda binary might
# reside in (or mamba, if it is the preferred one).
_CONDA_COMMAND = os.environ.get("CONDA_EXE", "conda")
_FAL_CONDA_HOME = os.getenv("FAL_CONDA_HOME")


@dataclass
class CondaEnvironment(BaseEnvironment[Path], make_thread_safe=True):
    packages: List[str]
    inherit_from_local: bool = False

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> CondaEnvironment:
        user_provided_packages = config.get("packages", [])
        # TODO: remove this once cross-Python-version serialization is
        # working.
        for raw_requirement in user_provided_packages:
            raw_requirement = raw_requirement.replace(" ", "")
            if raw_requirement.startswith("python"):
                continue

            # Ensure that the package is not python-something but rather
            # python followed by any of the version constraints.
            version_identifier = raw_requirement[len("python") :]
            if version_identifier and version_identifier[0] in (
                "=",
                "<",
                ">",
                "!",
            ):
                raise RuntimeError(
                    "Conda environments cannot customize their Python version."
                )

        # We currently don't support sending/receiving data between
        # different Python versions so we need to make sure that the
        # conda environment that is created also uses the same Python version.
        python_version = sysconfig.get_python_version()
        final_packages = user_provided_packages + [f"python={python_version}"]

        inherit_from_local = config.get("_inherit_from_local", False)
        return cls(final_packages, inherit_from_local=inherit_from_local)

    @property
    def key(self) -> str:
        return hashlib.sha256(" ".join(self.packages).encode()).hexdigest()

    def _get_or_create(self) -> Path:
        env_path = _BASE_CONDA_DIR / self.key
        if env_path.exists():
            return env_path

        with rmdir_on_fail(env_path):
            self._run_conda(
                "create",
                "--yes",
                # The environment will be created under $BASE_CACHE_DIR/conda
                # so that in the future we can reuse it.
                "--prefix",
                env_path,
                *self.packages,
            )
        return env_path

    def _run_conda(self, *args, **kwargs) -> None:
        log_env(self, "Installing conda environment")
        conda_executable = get_conda_executable()
        subprocess.check_call([conda_executable, *args], **kwargs, text=True)

    def open_connection(self, conn_info: Path) -> DualPythonIPC:
        if self.inherit_from_local:
            # Instead of creating a separate environment that only has
            # the same versions of fal/dbt-core etc. you have locally,
            # we can also use your environment as the primary. This is
            # mainly for the development time where the fal or dbt-core
            # you are using is not available on PyPI yet.
            primary_env_path = Path(sys.exec_prefix)
        else:
            primary_env = get_primary_virtual_env()
            primary_env_path = primary_env.get_or_create()
        secondary_env_path = conn_info
        return DualPythonIPC(self, primary_env_path, secondary_env_path)


@cache_static
def get_conda_executable() -> Path:
    for path in [_FAL_CONDA_HOME, None]:
        conda_path = shutil.which(_CONDA_COMMAND, path=path)
        if conda_path is not None:
            return conda_path
    else:
        # TODO: we should probably point to the instructions on how you
        # can install conda here.
        raise RuntimeError(
            "Could not find conda executable. Please install conda or set FAL_CONDA_HOME."
        )



from __future__ import annotations

from pathlib import Path
from typing import Iterator, List, Optional, Tuple

import importlib_metadata

from fal.dbt.utils import cache_static

import importlib_metadata


def _get_project_root_path(pacakge: str) -> Path:
    import fal.dbt as fal

    # If this is a development version, we'll install
    # the current fal itself.
    path = Path(fal.__file__)
    while path is not None:
        if (path.parent / ".git").exists():
            break
        path = path.parent
    return path / pacakge


def _get_dbt_packages() -> Iterator[Tuple[str, Optional[str]]]:
    # package_distributions will return a mapping of top-level package names to a list of distribution names (
    # the PyPI names instead of the import names). An example distribution info is the following, which
    # contains both the main exporter of the top-level name (dbt-core) as well as all the packages that
    # export anything to that namespace:
    #   {"dbt": ["dbt-core", "dbt-postgres", "dbt-athena-adapter"]}
    #
    # This won't only include dbt.adapters.xxx, but anything that might export anything to the dbt namespace
    # (e.g. a hypothetical plugin that only exports stuff to dbt.includes.xxx) which in theory would allow us
    # to replicate the exact environment.
    package_distributions = importlib_metadata.packages_distributions()
    for dbt_plugin_name in package_distributions.get("dbt", []):
        distribution = importlib_metadata.distribution(dbt_plugin_name)

        # Skip dbt-core since it will be determined by other packages being installed
        if dbt_plugin_name == "dbt-core":
            continue

        # Skip dbt-fal since it is already handled by _get_dbt_fal_package_name
        if dbt_plugin_name == "dbt-fal":
            continue

        yield dbt_plugin_name, distribution.version


def _find_fal_extras(package: str) -> set[str]:
    # Return a possible set of extras that might be required when installing
    # fal in the new environment. The original form which the user has installed
    # is not present to us (it is not saved anywhere during the package installation
    # process, so there is no way for us to know how a user initially installed fal).
    # We'll therefore be as generous as possible and install all the extras for all
    # the dbt.adapters that the user currently has (so this will still be a subset
    # of dependencies, e.g. if there is no dbt.adapter.duckdb then we won't include
    # duckdb as an extra).

    import pkgutil

    import dbt.adapters

    fal_dist = importlib_metadata.distribution(package)
    all_extras = fal_dist.metadata.get_all("Provides-Extra", [])

    # This list is different from the one we obtain in _get_dbt_packages
    # since the names here are the actual import names, not the PyPI names
    # (e.g. this one will say athena, and the other one will say dbt-athena-adapter).
    available_dbt_adapters = {
        module_info.name
        for module_info in pkgutil.iter_modules(dbt.adapters.__path__)
        if module_info.ispkg
    }

    # There will be adapters which we won't have an extra for (e.g. oraceledb)
    # and there will be extras which the user did not install the adapter for
    # (e.g. dbt-redshift). We want to take the intersection of all the adapters
    # that the user has installed and all the extras that fal provides and find
    # the smallest possible subset of extras that we can install.
    return available_dbt_adapters.intersection(all_extras)

def _running_pre_release() -> bool:
    raw_fal_version = importlib_metadata.version("dbt-fal")
    return _version_is_prerelease(raw_fal_version)

def _version_is_prerelease(raw_version: str) -> bool:
    from packaging.version import Version

    package_version = Version(raw_version)
    return package_version.is_prerelease

def _get_dbt_fal_package() -> Tuple[str, Optional[str]]:
    if _running_pre_release():
        proj_path = _get_project_root_path("adapter")
        if proj_path.exists():
            # We are going to install it from the local path.
            dbt_fal_dep = str(proj_path)
            dbt_fal_version = None
        else:
            # We are going to install it from PyPI.
            dbt_fal_dep = "dbt-fal"
            try:
                dbt_fal_version = importlib_metadata.version("dbt-fal")
            except importlib_metadata.PackageNotFoundError:
                # TODO: remove once `fal` is no longer a supported package
                dbt_fal_version = importlib_metadata.version("fal")
    else:
        dbt_fal_dep = "dbt-fal"
        try:
            dbt_fal_version = importlib_metadata.version("dbt-fal")
        except importlib_metadata.PackageNotFoundError:
            # TODO: remove once `fal` is no longer a supported package
            dbt_fal_version = importlib_metadata.version("fal")

    try:
        dbt_fal_extras = _find_fal_extras("dbt-fal")
    except importlib_metadata.PackageNotFoundError:
        # TODO: remove once `fal` is no longer a supported package
        dbt_fal_extras = _find_fal_extras("fal")

    if dbt_fal_extras:
        dbt_fal_dep += f"[{','.join(dbt_fal_extras)}]"

    return dbt_fal_dep, dbt_fal_version

def get_default_requirements() -> Iterator[Tuple[str, Optional[str]]]:
    yield _get_dbt_fal_package()
    yield from _get_dbt_packages()


@cache_static
def get_default_pip_dependencies() -> List[str]:
    return [
        f"{package}=={version}" if version else package
        for package, version in get_default_requirements()
    ]



from typing import Optional

from collections import defaultdict
from contextlib import contextmanager
from dbt.adapters.base.impl import BaseAdapter
from dbt.adapters.base.relation import BaseRelation
from dbt.adapters.protocol import AdapterConfig
from dbt.adapters.factory import FACTORY

# TODO: offer in `from isolate import is_agent`
from isolate.connections.common import is_agent

from .connections import FalEncCredentials
from .wrappers import FalEncAdapterWrapper, FalCredentialsWrapper

from .load_db_profile import load_profiles_info_1_5


class FalConfigs(AdapterConfig):
    fal_environment: Optional[str]
    fal_machine: Optional[str]


@contextmanager
def _release_plugin_lock():
    FACTORY.lock.release()
    try:
        yield
    finally:
        FACTORY.lock.acquire()


DB_PROFILE = None
DB_RELATION = BaseRelation
OVERRIDE_PROPERTIES = {}

# NOTE: Should this file run on isolate agents? Could we skip it entirely and build a FalEncAdapterWrapper directly?
if not is_agent():
    DB_PROFILE, OVERRIDE_PROPERTIES = load_profiles_info_1_5()
    DB_RELATION = FACTORY.get_relation_class_by_name(DB_PROFILE.credentials.type)


class FalEncAdapter(BaseAdapter):
    Relation = DB_RELATION  # type: ignore

    # TODO: how do we actually use this?
    AdapterSpecificConfigs = FalConfigs

    def __new__(cls, config):
        # There are two different credentials types which can be passed to FalEncAdapter
        # 1. FalEncCredentials
        # 2. FalCredentialsWrapper
        #
        # For `FalEncCredentials`, we have to go through parsing the profiles.yml (so that we
        # can obtain the real 'db' credentials). But for the other one, we can just use
        # the bound credentials directly (e.g. in isolated mode where we don't actually
        # have access to the profiles.yml file).

        fal_credentials = config.credentials
        if isinstance(fal_credentials, FalEncCredentials):
            db_credentials = DB_PROFILE.credentials
        else:
            # Since profile construction (in the case above) already registers the
            # adapter plugin for the db type, we need to also mimic that here.
            assert isinstance(fal_credentials, FalCredentialsWrapper)
            db_credentials = fal_credentials._db_creds
            with _release_plugin_lock():
                FACTORY.load_plugin(db_credentials.type)

        # TODO: maybe we can do this better?
        with _release_plugin_lock():
            db_adapter_class = FACTORY.get_adapter_class_by_name(db_credentials.type)
            original_plugin = FACTORY.get_plugin_by_name(fal_credentials.type)
            original_plugin.dependencies = [db_credentials.type]

        config.python_adapter_credentials = fal_credentials
        config.sql_adapter_credentials = db_credentials

        for key in OVERRIDE_PROPERTIES:
            if OVERRIDE_PROPERTIES[key] is not None:
                setattr(config, key, OVERRIDE_PROPERTIES[key])

        with _release_plugin_lock():
            # Temporary credentials for register
            config.credentials = config.sql_adapter_credentials
            FACTORY.register_adapter(config)
            config.credentials = FalCredentialsWrapper(config.sql_adapter_credentials)

        return FalEncAdapterWrapper(db_adapter_class, config)

    @classmethod
    def type(cls):
        return "fal"



from contextlib import contextmanager
from typing import Optional

from dbt.exceptions import DbtRuntimeError
from dbt.adapters.base.relation import BaseRelation
from dbt.adapters.base.impl import BaseAdapter

from dbt.adapters.fal_experimental.connections import (
    TeleportCredentials,
    TeleportTypeEnum,
)

from dbt.fal.adapters.teleport.impl import TeleportAdapter
from dbt.fal.adapters.teleport.info import TeleportInfo


class DuckDBAdapterTeleport(TeleportAdapter):
    def __init__(
        self, db_adapter: BaseAdapter, teleport_credentials: TeleportCredentials
    ):
        self._db_adapter = db_adapter
        self.credentials = teleport_credentials
        with self._db_adapter.connection_named("teleport:init"):
            self._db_adapter.execute("INSTALL parquet")
            self._db_adapter.execute("INSTALL httpfs")

    @classmethod
    def storage_formats(cls):
        return ["parquet"]

    def teleport_from_external_storage(
        self, relation: BaseRelation, relation_path: str, teleport_info: TeleportInfo
    ):
        assert (
            teleport_info.format == "parquet"
        ), "duckdb only supports parquet format for Teleport"

        url = teleport_info.build_url(relation_path)

        with self._db_adapter.connection_named("teleport:copy_from"):
            if self.credentials.type == TeleportTypeEnum.LOCAL:
                rendered_macro = self._db_adapter.execute_macro(
                    "duckdb__copy_from_parquet",
                    kwargs={"relation": relation, "url": url},
                )
                self._db_adapter.execute(rendered_macro)

            elif self.credentials.type == TeleportTypeEnum.REMOTE_S3:
                with self._s3_setup():
                    rendered_macro = self._db_adapter.execute_macro(
                        "duckdb__copy_from_parquet",
                        kwargs={"relation": relation, "url": url},
                    )
                    self._db_adapter.execute(rendered_macro)
            else:
                raise RuntimeError(
                    f"Teleport type {self.credentials.type} not supported"
                )

    def teleport_to_external_storage(
        self, relation: BaseRelation, teleport_info: TeleportInfo
    ):
        assert (
            teleport_info.format == "parquet"
        ), "duckdb only supports parquet format for Teleport"

        rel_path = teleport_info.build_relation_path(relation)
        url = teleport_info.build_url(rel_path)

        with self._db_adapter.connection_named("teleport:copy_to"):
            if self.credentials.type == TeleportTypeEnum.LOCAL:
                rendered_macro = self._db_adapter.execute_macro(
                    "duckdb__copy_to", kwargs={"relation": relation, "url": url}
                )
                self._db_adapter.execute(rendered_macro)
            elif self.credentials.type == TeleportTypeEnum.REMOTE_S3:
                with self._s3_setup():
                    rendered_macro = self._db_adapter.execute_macro(
                        "duckdb__copy_to", kwargs={"relation": relation, "url": url}
                    )
                    self._db_adapter.execute(rendered_macro)
            else:
                raise RuntimeError(
                    f"Teleport type {self.credentials.type} not supported"
                )

        return rel_path

    def _get_setting(self, name: str):
        try:
            _, table = self._db_adapter.execute(
                f"SELECT current_setting('{name}')", fetch=True
            )
            return table.rows[0][0]
        except DbtRuntimeError:
            return None

    def _set_setting(self, name: str, value: Optional[str]):
        if value:
            self._db_adapter.execute(f"SET {name} = '{value}'")
        else:
            # HACK while we get a response https://github.com/duckdb/duckdb/issues/4998
            self._db_adapter.execute(f"SET {name} = ''")

    @contextmanager
    def _s3_setup(self):
        self._db_adapter.execute("LOAD parquet")
        self._db_adapter.execute("LOAD httpfs")

        old_region = self._get_setting("s3_region")
        old_access_key_id = self._get_setting("s3_access_key_id")
        old_secret_access_key = self._get_setting("s3_secret_access_key")

        self._set_setting("s3_region", self.credentials.s3_region)
        self._set_setting("s3_access_key_id", self.credentials.s3_access_key_id)
        self._set_setting("s3_secret_access_key", self.credentials.s3_access_key)

        yield

        self._set_setting("s3_region", old_region)
        self._set_setting("s3_access_key_id", old_access_key_id)
        self._set_setting("s3_secret_access_key", old_secret_access_key)



from dbt.adapters.base import AdapterPlugin

from dbt.adapters.fal_experimental.connections import FalCredentials
from dbt.adapters.fal_experimental.impl import FalAdapter
from dbt.include import fal_experimental

Plugin = AdapterPlugin(
    adapter=FalAdapter,
    credentials=FalCredentials,
    include_path=fal_experimental.PACKAGE_PATH,
)
